"""Tests for field groups — all-or-nothing validation during merge."""

from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent

import pytest

from dature import FieldGroup, FieldMergeStrategy, LoadMetadata, MergeMetadata, MergeRule, MergeStrategy, load
from dature.errors import FieldGroupError
from dature.field_path import F


class TestFieldGroupAllChanged:
    def test_all_fields_changed_last_wins(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote", "port": 9090}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
                strategy=MergeStrategy.LAST_WINS,
                field_groups=(FieldGroup(F[Config].host, F[Config].port),),
            ),
            Config,
        )

        assert result.host == "remote"
        assert result.port == 9090

    def test_all_fields_changed_first_wins(self, tmp_path: Path):
        first = tmp_path / "first.json"
        first.write_text('{"host": "first-host", "port": 1000}')

        second = tmp_path / "second.json"
        second.write_text('{"host": "second-host", "port": 2000}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(first)),
                    LoadMetadata(file_=str(second)),
                ),
                strategy=MergeStrategy.FIRST_WINS,
                field_groups=(FieldGroup(F[Config].host, F[Config].port),),
            ),
            Config,
        )

        assert result.host == "first-host"
        assert result.port == 1000


class TestFieldGroupNoneChanged:
    def test_no_fields_changed(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "localhost", "port": 3000}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
                field_groups=(FieldGroup(F[Config].host, F[Config].port),),
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 3000

    def test_source_missing_all_group_fields(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000, "debug": false}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"debug": true}')

        @dataclass
        class Config:
            host: str
            port: int
            debug: bool

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
                field_groups=(FieldGroup(F[Config].host, F[Config].port),),
            ),
            Config,
        )

        assert result.host == "localhost"
        assert result.port == 3000
        assert result.debug is True


class TestFieldGroupPartialChange:
    def test_partial_change_raises(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote"}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(FieldGroup(F[Config].host, F[Config].port),),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (host, port) partially overridden in source 1
                changed:   host (from source {overrides_meta!r})
                unchanged: port (from source {defaults_meta!r})
            """)

    def test_partial_change_field_present_but_equal(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote", "port": 3000}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(FieldGroup(F[Config].host, F[Config].port),),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (host, port) partially overridden in source 1
                changed:   host (from source {overrides_meta!r})
                unchanged: port (from source {defaults_meta!r})
            """)

    def test_partial_change_with_first_wins(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote"}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(FieldGroupError):
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=str(defaults)),
                        LoadMetadata(file_=str(overrides)),
                    ),
                    strategy=MergeStrategy.FIRST_WINS,
                    field_groups=(FieldGroup(F[Config].host, F[Config].port),),
                ),
                Config,
            )

    def test_partial_change_with_raise_on_conflict(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote"}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(FieldGroupError):
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=str(defaults)),
                        LoadMetadata(file_=str(overrides)),
                    ),
                    strategy=MergeStrategy.RAISE_ON_CONFLICT,
                    field_groups=(FieldGroup(F[Config].host, F[Config].port),),
                ),
                Config,
            )


class TestFieldGroupAutoExpand:
    def test_auto_expand_nested_dataclass(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"database": {"host": "localhost", "port": 5432}}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"database": {"host": "remote"}}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(FieldGroup(F[Config].database),),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (database.host, database.port) partially overridden in source 1
                changed:   database.host (from source {overrides_meta!r})
                unchanged: database.port (from source {defaults_meta!r})
            """)

    def test_auto_expand_all_changed_ok(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"database": {"host": "localhost", "port": 5432}}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"database": {"host": "remote", "port": 3306}}')

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
                field_groups=(FieldGroup(F[Config].database),),
            ),
            Config,
        )

        assert result.database.host == "remote"
        assert result.database.port == 3306


class TestFieldGroupThreeSources:
    def test_three_sources_violation_on_second(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{"host": "a-host", "port": 1000}')

        b = tmp_path / "b.json"
        b.write_text('{"host": "b-host"}')

        c = tmp_path / "c.json"
        c.write_text('{"host": "c-host", "port": 3000}')

        a_meta = LoadMetadata(file_=str(a))
        b_meta = LoadMetadata(file_=str(b))
        c_meta = LoadMetadata(file_=str(c))

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(a_meta, b_meta, c_meta),
                    field_groups=(FieldGroup(F[Config].host, F[Config].port),),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (host, port) partially overridden in source 1
                changed:   host (from source {b_meta!r})
                unchanged: port (from source {a_meta!r})
            """)

    def test_three_sources_all_ok(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{"host": "a-host", "port": 1000}')

        b = tmp_path / "b.json"
        b.write_text('{"host": "b-host", "port": 2000}')

        c = tmp_path / "c.json"
        c.write_text('{"host": "c-host", "port": 3000}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(a)),
                    LoadMetadata(file_=str(b)),
                    LoadMetadata(file_=str(c)),
                ),
                field_groups=(FieldGroup(F[Config].host, F[Config].port),),
            ),
            Config,
        )

        assert result.host == "c-host"
        assert result.port == 3000


class TestFieldGroupMultipleGroups:
    def test_one_ok_one_violated(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000, "user": "admin", "password": "secret"}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote", "port": 9090, "user": "root"}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))

        @dataclass
        class Config:
            host: str
            port: int
            user: str
            password: str

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(
                        FieldGroup(F[Config].host, F[Config].port),
                        FieldGroup(F[Config].user, F[Config].password),
                    ),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (user, password) partially overridden in source 1
                changed:   user (from source {overrides_meta!r})
                unchanged: password (from source {defaults_meta!r})
            """)


class TestFieldGroupWithFieldMerges:
    def test_compatible_with_field_merges(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000, "tags": ["a"]}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote", "port": 9090, "tags": ["b"]}')

        @dataclass
        class Config:
            host: str
            port: int
            tags: list[str]

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
                field_merges=(MergeRule(F[Config].tags, FieldMergeStrategy.APPEND),),
                field_groups=(FieldGroup(F[Config].host, F[Config].port),),
            ),
            Config,
        )

        assert result.host == "remote"
        assert result.port == 9090
        assert result.tags == ["a", "b"]


class TestFieldGroupDecorator:
    def test_decorator_with_field_groups(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote", "port": 9090}')

        meta = MergeMetadata(
            sources=(
                LoadMetadata(file_=str(defaults)),
                LoadMetadata(file_=str(overrides)),
            ),
            field_groups=(FieldGroup(F["Config"].host, F["Config"].port),),
        )

        @load(meta)
        @dataclass
        class Config:
            host: str
            port: int

        config = Config()
        assert config.host == "remote"
        assert config.port == 9090

    def test_decorator_partial_change_raises(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote"}')

        meta = MergeMetadata(
            sources=(
                LoadMetadata(file_=str(defaults)),
                LoadMetadata(file_=str(overrides)),
            ),
            field_groups=(FieldGroup(F["Config"].host, F["Config"].port),),
        )

        @load(meta)
        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(FieldGroupError):
            Config()


class TestFieldGroupErrorFormat:
    def test_error_message_format(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000, "debug": false}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote", "debug": true}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))

        @dataclass
        class Config:
            host: str
            port: int
            debug: bool

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(FieldGroup(F[Config].host, F[Config].port),),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (host, port) partially overridden in source 1
                changed:   host (from source {overrides_meta!r})
                unchanged: port (from source {defaults_meta!r})
            """)

    def test_multiple_violations_message(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000, "user": "admin", "password": "secret"}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"host": "remote", "user": "root"}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))

        @dataclass
        class Config:
            host: str
            port: int
            user: str
            password: str

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(
                        FieldGroup(F[Config].host, F[Config].port),
                        FieldGroup(F[Config].user, F[Config].password),
                    ),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (2)

              Field group (host, port) partially overridden in source 1
                changed:   host (from source {overrides_meta!r})
                unchanged: port (from source {defaults_meta!r})

              Field group (user, password) partially overridden in source 1
                changed:   user (from source {overrides_meta!r})
                unchanged: password (from source {defaults_meta!r})
            """)


class TestFieldGroupMixedExpandAndFlat:
    def test_all_changed_ok(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text(
            '{"database": {"host": "localhost", "port": 5432}, "timeout": 30}',
        )

        overrides = tmp_path / "overrides.json"
        overrides.write_text(
            '{"database": {"host": "remote", "port": 3306}, "timeout": 60}',
        )

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database
            timeout: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
                field_groups=(FieldGroup(F[Config].database, F[Config].timeout),),
            ),
            Config,
        )

        assert result.database.host == "remote"
        assert result.database.port == 3306
        assert result.timeout == 60

    def test_none_changed_ok(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text(
            '{"database": {"host": "localhost", "port": 5432}, "timeout": 30}',
        )

        overrides = tmp_path / "overrides.json"
        overrides.write_text(
            '{"database": {"host": "localhost", "port": 5432}, "timeout": 30}',
        )

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database
            timeout: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
                field_groups=(FieldGroup(F[Config].database, F[Config].timeout),),
            ),
            Config,
        )

        assert result.database.host == "localhost"
        assert result.database.port == 5432
        assert result.timeout == 30

    def test_flat_changed_nested_not(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text(
            '{"database": {"host": "localhost", "port": 5432}, "timeout": 30}',
        )

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"timeout": 60}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database
            timeout: int

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(FieldGroup(F[Config].database, F[Config].timeout),),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (database.host, database.port, timeout) partially overridden in source 1
                changed:   timeout (from source {overrides_meta!r})
                unchanged: database.host (from source {defaults_meta!r}), database.port (from source {defaults_meta!r})
            """)

    def test_nested_partial_flat_not(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text(
            '{"database": {"host": "localhost", "port": 5432}, "timeout": 30}',
        )

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"database": {"host": "remote"}}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database
            timeout: int

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(FieldGroup(F[Config].database, F[Config].timeout),),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (database.host, database.port, timeout) partially overridden in source 1
                changed:   database.host (from source {overrides_meta!r})
                unchanged: database.port (from source {defaults_meta!r}), timeout (from source {defaults_meta!r})
            """)

    def test_nested_all_changed_flat_not(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text(
            '{"database": {"host": "localhost", "port": 5432}, "timeout": 30}',
        )

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"database": {"host": "remote", "port": 3306}}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))
        d = repr(defaults_meta)
        o = repr(overrides_meta)

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database
            timeout: int

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(FieldGroup(F[Config].database, F[Config].timeout),),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (database.host, database.port, timeout) partially overridden in source 1
                changed:   database.host (from source {o}), database.port (from source {o})
                unchanged: timeout (from source {d})
            """)


class TestFieldGroupSameFieldNameNested:
    def test_all_changed_ok(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text(
            '{"user_name": "root-old", "inner": {"user_name": "nested-old"}}',
        )

        overrides = tmp_path / "overrides.json"
        overrides.write_text(
            '{"user_name": "root-new", "inner": {"user_name": "nested-new"}}',
        )

        @dataclass
        class Inner:
            user_name: str

        @dataclass
        class Config:
            user_name: str
            inner: Inner

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
                field_groups=(FieldGroup(F[Config].user_name, F[Config].inner.user_name),),
            ),
            Config,
        )

        assert result.user_name == "root-new"
        assert result.inner.user_name == "nested-new"

    def test_only_root_changed_raises(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text(
            '{"user_name": "root-old", "inner": {"user_name": "nested-old"}}',
        )

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"user_name": "root-new"}')

        defaults_meta = LoadMetadata(file_=str(defaults))
        overrides_meta = LoadMetadata(file_=str(overrides))

        @dataclass
        class Inner:
            user_name: str

        @dataclass
        class Config:
            user_name: str
            inner: Inner

        with pytest.raises(FieldGroupError) as exc_info:
            load(
                MergeMetadata(
                    sources=(defaults_meta, overrides_meta),
                    field_groups=(FieldGroup(F[Config].user_name, F[Config].inner.user_name),),
                ),
                Config,
            )

        assert str(exc_info.value) == dedent(f"""\
            Config field group errors (1)

              Field group (user_name, inner.user_name) partially overridden in source 1
                changed:   user_name (from source {overrides_meta!r})
                unchanged: inner.user_name (from source {defaults_meta!r})
            """)



"""Tests for main.py — public load() API."""

from dataclasses import dataclass, field
from pathlib import Path

import pytest

from dature import LoadMetadata, load
from dature.sources_loader.env_ import EnvFileLoader
from dature.sources_loader.ini_ import IniLoader
from dature.sources_loader.json5_ import Json5Loader
from dature.sources_loader.json_ import JsonLoader
from dature.sources_loader.toml_ import TomlLoader
from dature.sources_loader.yaml_ import Yaml11Loader, Yaml12Loader


def _all_file_loaders() -> list[type]:
    return [EnvFileLoader, Yaml11Loader, Yaml12Loader, JsonLoader, Json5Loader, TomlLoader, IniLoader]


class TestLoadAsDecorator:
    def test_loads_from_file(self, tmp_path: Path) -> None:
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "FromFile", "port": 8080}')

        metadata = LoadMetadata(file_=str(json_file))

        @load(metadata)
        @dataclass
        class Config:
            name: str
            port: int

        config = Config()
        assert config.name == "FromFile"
        assert config.port == 8080

    def test_loads_from_env(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("APP_NAME", "EnvApp")
        monkeypatch.setenv("APP_PORT", "3000")

        metadata = LoadMetadata(prefix="APP_")

        @load(metadata)
        @dataclass
        class Config:
            name: str
            port: int

        config = Config()
        assert config.name == "EnvApp"
        assert config.port == 3000

    def test_default_metadata(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("MY_VAR", "test_value")

        @load()
        @dataclass
        class Config:
            my_var: str

        config = Config()
        assert config.my_var == "test_value"

    def test_explicit_loader_overrides_extension(self, tmp_path: Path) -> None:
        txt_file = tmp_path / "config.txt"
        txt_file.write_text('{"app_name": "OverrideApp"}')

        metadata = LoadMetadata(file_=str(txt_file), loader=JsonLoader)

        @load(metadata)
        @dataclass
        class Config:
            app_name: str

        config = Config()
        assert config.app_name == "OverrideApp"

    def test_priority(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("LOADED_VAR", "loaded")
        monkeypatch.setenv("OVERRIDDEN_VAR", "loaded")

        @load()
        @dataclass
        class Config:
            overridden_var: str
            default_var: str = field(default="default")
            loaded_var: str = field(default="default")

        config = Config(overridden_var="from_init")

        assert config.default_var == "default"
        assert config.loaded_var == "loaded"
        assert config.overridden_var == "from_init"

    def test_invalid_decorator_order(self) -> None:
        with pytest.raises(TypeError, match="Config must be a dataclass"):

            @dataclass
            @load()
            class Config:
                pass


class TestCache:
    def test_cache_enabled_by_default(self, tmp_path: Path) -> None:
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "original", "port": 8080}')

        metadata = LoadMetadata(file_=str(json_file))

        @load(metadata)
        @dataclass
        class Config:
            name: str
            port: int

        first = Config()
        json_file.write_text('{"name": "updated", "port": 9090}')
        second = Config()

        assert first.name == "original"
        assert second.name == "original"

    def test_cache_disabled(self, tmp_path: Path) -> None:
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "original", "port": 8080}')

        metadata = LoadMetadata(file_=str(json_file))

        @load(metadata, cache=False)
        @dataclass
        class Config:
            name: str
            port: int

        first = Config()
        json_file.write_text('{"name": "updated", "port": 9090}')
        second = Config()

        assert first.name == "original"
        assert second.name == "updated"


class TestLoadAsFunction:
    def test_loads_from_file(self, tmp_path: Path) -> None:
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "FromFile", "port": 9090}')

        @dataclass
        class Config:
            name: str
            port: int

        metadata = LoadMetadata(file_=str(json_file))
        result = load(metadata, Config)

        assert result.name == "FromFile"
        assert result.port == 9090

    def test_loads_from_env(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("APP_NAME", "EnvFunc")
        monkeypatch.setenv("APP_DEBUG", "true")

        @dataclass
        class Config:
            name: str
            debug: bool

        metadata = LoadMetadata(prefix="APP_")
        result = load(metadata, Config)

        assert result.name == "EnvFunc"
        assert result.debug is True

    def test_default_metadata(self, monkeypatch: pytest.MonkeyPatch) -> None:
        monkeypatch.setenv("MY_VAR", "from_env")

        @dataclass
        class Config:
            my_var: str

        result = load(None, Config)

        assert result.my_var == "from_env"


class TestFileNotFoundWithLoad:
    @pytest.mark.parametrize(
        "loader_class",
        _all_file_loaders(),
    )
    def test_load_function_single_source_file_not_found(self, loader_class: type) -> None:

        @dataclass
        class Config:
            name: str

        metadata = LoadMetadata(file_="/non/existent/file.json", loader=loader_class)

        with pytest.raises(FileNotFoundError):
            load(metadata, Config)

    @pytest.mark.parametrize(
        "loader_class",
        _all_file_loaders(),
    )
    def test_load_decorator_single_source_file_not_found(self, loader_class: type) -> None:
        metadata = LoadMetadata(file_="/non/existent/config.json", loader=loader_class)

        @load(metadata)
        @dataclass
        class Config:
            name: str

        with pytest.raises(FileNotFoundError):
            Config()



import base64
import re
from datetime import timedelta
from urllib.parse import urlparse

from dature.fields import ByteSize, PaymentCardNumber, SecretStr
from dature.types import URL, Base64UrlBytes, Base64UrlStr

_TIMEDELTA_WITH_DAYS_RE = re.compile(
    r"^(-?\d+)\s+days?,\s*(\d+):(\d{2}):(\d{2})(?:\.(\d+))?$",
)
_TIMEDELTA_TIME_ONLY_RE = re.compile(
    r"^(\d+):(\d{2}):(\d{2})(?:\.(\d+))?$",
)


def bytes_from_string(value: str) -> bytes:
    return value.encode("utf-8")


def complex_from_string(value: str) -> complex:
    return complex(value.replace(" ", ""))


def timedelta_from_string(value: str) -> timedelta:
    """Parse str(timedelta(...)) format: '1 day, 2:30:00', '0:45:00', '-1 day, 23:59:59'."""
    if match := _TIMEDELTA_WITH_DAYS_RE.match(value):
        days = int(match.group(1))
        hours = int(match.group(2))
        minutes = int(match.group(3))
        seconds = int(match.group(4))
        microseconds = int(match.group(5).ljust(6, "0")) if match.group(5) is not None else 0
        return timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds, microseconds=microseconds)

    if match := _TIMEDELTA_TIME_ONLY_RE.match(value):
        hours = int(match.group(1))
        minutes = int(match.group(2))
        seconds = int(match.group(3))
        microseconds = int(match.group(4).ljust(6, "0")) if match.group(4) is not None else 0
        return timedelta(hours=hours, minutes=minutes, seconds=seconds, microseconds=microseconds)

    msg = f"Invalid timedelta format: {value!r}"
    raise ValueError(msg)


def url_from_string(value: str) -> URL:
    return urlparse(value)


def base64url_bytes_from_string(value: str) -> Base64UrlBytes:
    return base64.urlsafe_b64decode(value)


def base64url_str_from_string(value: str) -> Base64UrlStr:
    return base64.urlsafe_b64decode(value).decode("utf-8")


def secret_str_from_string(value: str) -> SecretStr:
    return SecretStr(value)


def payment_card_number_from_string(value: str) -> PaymentCardNumber:
    return PaymentCardNumber(value)


def byte_size_from_string(value: str | int) -> ByteSize:
    return ByteSize(value)



from collections.abc import Callable
from dataclasses import dataclass


@dataclass(frozen=True, slots=True, kw_only=True)
class RootValidator:
    func: Callable[..., bool]
    error_message: str = "Root validation failed"

    def get_validator_func(self) -> Callable[..., bool]:
        return self.func

    def get_error_message(self) -> str:
        return self.error_message



"""Tests for LoadReport and debug logging."""

import logging
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Annotated

import pytest

from dature import LoadMetadata, MergeMetadata, MergeStrategy, get_load_report, load
from dature.errors import DatureConfigError
from dature.load_report import FieldOrigin, LoadReport, SourceEntry
from dature.validators.number import Ge


class TestGetLoadReportMergeFunction:
    def test_last_wins(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
            ),
            Config,
            debug=True,
        )

        report = get_load_report(result)

        expected = LoadReport(
            dataclass_name="Config",
            strategy=MergeStrategy.LAST_WINS,
            sources=(
                SourceEntry(
                    index=0,
                    file_path=str(defaults),
                    loader_type="json",
                    raw_data={"host": "localhost", "port": 3000},
                ),
                SourceEntry(
                    index=1,
                    file_path=str(overrides),
                    loader_type="json",
                    raw_data={"port": 8080},
                ),
            ),
            field_origins=(
                FieldOrigin(
                    key="host",
                    value="localhost",
                    source_index=0,
                    source_file=str(defaults),
                    source_loader_type="json",
                ),
                FieldOrigin(
                    key="port",
                    value=8080,
                    source_index=1,
                    source_file=str(overrides),
                    source_loader_type="json",
                ),
            ),
            merged_data={"host": "localhost", "port": 8080},
        )
        assert expected == report

    def test_first_wins(self, tmp_path: Path):
        first = tmp_path / "first.json"
        first.write_text('{"host": "first-host", "port": 1000}')

        second = tmp_path / "second.json"
        second.write_text('{"host": "second-host", "port": 2000}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(first)),
                    LoadMetadata(file_=str(second)),
                ),
                strategy=MergeStrategy.FIRST_WINS,
            ),
            Config,
            debug=True,
        )

        report = get_load_report(result)

        expected = LoadReport(
            dataclass_name="Config",
            strategy=MergeStrategy.FIRST_WINS,
            sources=(
                SourceEntry(
                    index=0,
                    file_path=str(first),
                    loader_type="json",
                    raw_data={"host": "first-host", "port": 1000},
                ),
                SourceEntry(
                    index=1,
                    file_path=str(second),
                    loader_type="json",
                    raw_data={"host": "second-host", "port": 2000},
                ),
            ),
            field_origins=(
                FieldOrigin(
                    key="host",
                    value="second-host",
                    source_index=0,
                    source_file=str(first),
                    source_loader_type="json",
                ),
                FieldOrigin(
                    key="port",
                    value=2000,
                    source_index=0,
                    source_file=str(first),
                    source_loader_type="json",
                ),
            ),
            merged_data={"host": "first-host", "port": 1000},
        )
        assert expected == report

    def test_nested_field_origins(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"database": {"host": "localhost", "port": 5432}}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"database": {"host": "prod-host"}}')

        @dataclass
        class Database:
            host: str
            port: int

        @dataclass
        class Config:
            database: Database

        result = load(
            MergeMetadata(
                sources=(
                    LoadMetadata(file_=str(defaults)),
                    LoadMetadata(file_=str(overrides)),
                ),
            ),
            Config,
            debug=True,
        )

        report = get_load_report(result)
        assert report is not None

        expected_origins = (
            FieldOrigin(
                key="database.host",
                value="prod-host",
                source_index=1,
                source_file=str(overrides),
                source_loader_type="json",
            ),
            FieldOrigin(
                key="database.port",
                value=5432,
                source_index=0,
                source_file=str(defaults),
                source_loader_type="json",
            ),
        )
        assert expected_origins == report.field_origins


class TestGetLoadReportSingleSource:
    def test_single_source_function(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test", "port": 8080}')

        @dataclass
        class Config:
            name: str
            port: int

        result = load(LoadMetadata(file_=str(json_file)), Config, debug=True)

        report = get_load_report(result)

        expected = LoadReport(
            dataclass_name="Config",
            strategy=None,
            sources=(
                SourceEntry(
                    index=0,
                    file_path=str(json_file),
                    loader_type="json",
                    raw_data={"name": "test", "port": 8080},
                ),
            ),
            field_origins=(
                FieldOrigin(
                    key="name",
                    value="test",
                    source_index=0,
                    source_file=str(json_file),
                    source_loader_type="json",
                ),
                FieldOrigin(
                    key="port",
                    value=8080,
                    source_index=0,
                    source_file=str(json_file),
                    source_loader_type="json",
                ),
            ),
            merged_data={"name": "test", "port": 8080},
        )
        assert expected == report


class TestGetLoadReportDecorator:
    def test_merge_decorator(self, tmp_path: Path):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"port": 9090}')

        meta = MergeMetadata(
            sources=(
                LoadMetadata(file_=str(defaults)),
                LoadMetadata(file_=str(overrides)),
            ),
        )

        @load(meta, debug=True)
        @dataclass
        class Config:
            host: str
            port: int

        config = Config()
        report = get_load_report(config)
        assert report is not None
        assert report.strategy == MergeStrategy.LAST_WINS
        assert len(report.sources) == 2

    def test_single_source_decorator(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": 3000}')

        @load(LoadMetadata(file_=str(json_file)), debug=True)
        @dataclass
        class Config:
            host: str
            port: int

        config = Config()
        report = get_load_report(config)
        assert report is not None
        assert report.strategy is None
        assert len(report.sources) == 1


class TestGetLoadReportWithoutDebug:
    def test_returns_none_with_warning(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            result = get_load_report("hello")

        assert result is None
        assert len(w) == 1
        assert "debug=True" in str(w[0].message)

    def test_no_report_without_debug(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": 3000}')

        @dataclass
        class Config:
            host: str
            port: int

        result = load(LoadMetadata(file_=str(json_file)), Config)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            report = get_load_report(result)

        assert report is None
        assert len(w) == 1
        assert "debug=True" in str(w[0].message)


class TestDebugLogging:
    def test_merge_debug_logs(self, tmp_path: Path, caplog: pytest.LogCaptureFixture):
        defaults = tmp_path / "defaults.json"
        defaults.write_text('{"host": "localhost", "port": 3000}')

        overrides = tmp_path / "overrides.json"
        overrides.write_text('{"port": 8080}')

        @dataclass
        class Config:
            host: str
            port: int

        with caplog.at_level(logging.DEBUG, logger="dature"):
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=str(defaults)),
                        LoadMetadata(file_=str(overrides)),
                    ),
                ),
                Config,
            )

        messages = [r.message for r in caplog.records if r.name == "dature"]

        expected = [
            f"[JsonLoader] load_raw: path={defaults},"
            " raw_keys=['host', 'port'], after_preprocessing_keys=['host', 'port']",
            f"[Config] Source 0 loaded: loader=json, file={defaults}, keys=['host', 'port']",
            "[Config] Source 0 raw data: {'host': 'localhost', 'port': 3000}",
            f"[JsonLoader] load_raw: path={overrides}, raw_keys=['port'], after_preprocessing_keys=['port']",
            f"[Config] Source 1 loaded: loader=json, file={overrides}, keys=['port']",
            "[Config] Source 1 raw data: {'port': 8080}",
            "[Config] Merge step 0 (strategy=last_wins): added=['host', 'port'], overwritten=[]",
            "[Config] State after step 0: {'host': 'localhost', 'port': 3000}",
            "[Config] Merge step 1 (strategy=last_wins): added=[], overwritten=['port']",
            "[Config] State after step 1: {'host': 'localhost', 'port': 8080}",
            "[Config] Merged result (strategy=last_wins, 2 sources): {'host': 'localhost', 'port': 8080}",
            f"[Config] Field 'host' = 'localhost'  <-- source 0 ({defaults})",
            f"[Config] Field 'port' = 8080  <-- source 1 ({overrides})",
        ]
        assert expected == messages

    def test_single_source_debug_logs(self, tmp_path: Path, caplog: pytest.LogCaptureFixture):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost", "port": 3000}')

        @dataclass
        class Config:
            host: str
            port: int

        with caplog.at_level(logging.DEBUG, logger="dature"):
            load(LoadMetadata(file_=str(json_file)), Config)

        messages = [r.message for r in caplog.records if r.name == "dature"]

        expected = [
            f"[JsonLoader] load_raw: path={json_file},"
            " raw_keys=['host', 'port'], after_preprocessing_keys=['host', 'port']",
            f"[Config] Single-source load: loader=json, file={json_file}",
            "[Config] Loaded data: {'host': 'localhost', 'port': 3000}",
        ]
        assert expected == messages


class TestLoadReportOnError:
    def test_merge_missing_field(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{"host": "localhost"}')

        b = tmp_path / "b.json"
        b.write_text('{"host": "override"}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError):
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=str(a)),
                        LoadMetadata(file_=str(b)),
                    ),
                ),
                Config,
                debug=True,
            )

        expected = LoadReport(
            dataclass_name="Config",
            strategy=MergeStrategy.LAST_WINS,
            sources=(
                SourceEntry(index=0, file_path=str(a), loader_type="json", raw_data={"host": "localhost"}),
                SourceEntry(index=1, file_path=str(b), loader_type="json", raw_data={"host": "override"}),
            ),
            field_origins=(
                FieldOrigin(
                    key="host",
                    value="override",
                    source_index=1,
                    source_file=str(b),
                    source_loader_type="json",
                ),
            ),
            merged_data={"host": "override"},
        )
        assert expected == get_load_report(Config)

    def test_merge_validation_error(self, tmp_path: Path):
        a = tmp_path / "a.json"
        a.write_text('{"port": -5}')

        b = tmp_path / "b.json"
        b.write_text('{"host": "localhost"}')

        @dataclass
        class Config:
            host: str
            port: Annotated[int, Ge(value=0)]

        with pytest.raises(DatureConfigError):
            load(
                MergeMetadata(
                    sources=(
                        LoadMetadata(file_=str(a)),
                        LoadMetadata(file_=str(b)),
                    ),
                ),
                Config,
                debug=True,
            )

        expected = LoadReport(
            dataclass_name="Config",
            strategy=MergeStrategy.LAST_WINS,
            sources=(
                SourceEntry(index=0, file_path=str(a), loader_type="json", raw_data={"port": -5}),
                SourceEntry(index=1, file_path=str(b), loader_type="json", raw_data={"host": "localhost"}),
            ),
            field_origins=(
                FieldOrigin(
                    key="host",
                    value="localhost",
                    source_index=1,
                    source_file=str(b),
                    source_loader_type="json",
                ),
                FieldOrigin(key="port", value=-5, source_index=0, source_file=str(a), source_loader_type="json"),
            ),
            merged_data={"host": "localhost", "port": -5},
        )
        assert expected == get_load_report(Config)

    def test_single_source_missing_field(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"host": "localhost"}')

        @dataclass
        class Config:
            host: str
            port: int

        with pytest.raises(DatureConfigError):
            load(LoadMetadata(file_=str(json_file)), Config, debug=True)

        expected = LoadReport(
            dataclass_name="Config",
            strategy=None,
            sources=(
                SourceEntry(index=0, file_path=str(json_file), loader_type="json", raw_data={"host": "localhost"}),
            ),
            field_origins=(
                FieldOrigin(
                    key="host",
                    value="localhost",
                    source_index=0,
                    source_file=str(json_file),
                    source_loader_type="json",
                ),
            ),
            merged_data={"host": "localhost"},
        )
        assert expected == get_load_report(Config)

    def test_single_source_validation_error(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"port": -1}')

        @dataclass
        class Config:
            port: Annotated[int, Ge(value=0)]

        with pytest.raises(DatureConfigError):
            load(LoadMetadata(file_=str(json_file)), Config, debug=True)

        expected = LoadReport(
            dataclass_name="Config",
            strategy=None,
            sources=(SourceEntry(index=0, file_path=str(json_file), loader_type="json", raw_data={"port": -1}),),
            field_origins=(
                FieldOrigin(
                    key="port",
                    value=-1,
                    source_index=0,
                    source_file=str(json_file),
                    source_loader_type="json",
                ),
            ),
            merged_data={"port": -1},
        )
        assert expected == get_load_report(Config)



from pathlib import Path
from typing import TYPE_CHECKING

from dature.sources_loader.env_ import EnvFileLoader, EnvLoader
from dature.sources_loader.ini_ import IniLoader
from dature.sources_loader.json_ import JsonLoader
from dature.sources_loader.toml_ import TomlLoader

if TYPE_CHECKING:
    from dature.protocols import LoaderProtocol

SUPPORTED_EXTENSIONS = (".cfg", ".env", ".ini", ".json", ".json5", ".toml", ".yaml", ".yml")


def _resolve_by_extension(extension: str) -> "type[LoaderProtocol]":
    match extension:
        case ".json":
            return JsonLoader
        case ".toml":
            return TomlLoader
        case ".ini" | ".cfg":
            return IniLoader
        case ".env":
            return EnvFileLoader
        case ".yaml" | ".yml":
            from dature.sources_loader.yaml_ import Yaml12Loader  # noqa: PLC0415

            return Yaml12Loader
        case ".json5":
            from dature.sources_loader.json5_ import Json5Loader  # noqa: PLC0415

            return Json5Loader
        case _:
            supported = ", ".join(SUPPORTED_EXTENSIONS)
            msg = (
                f"Cannot determine loader type for extension '{extension}'. "
                f"Please specify loader explicitly or use a supported extension: {supported}"
            )
            raise ValueError(msg)


def resolve_loader_class(
    loader: "type[LoaderProtocol] | None",
    file_: str | None,
) -> "type[LoaderProtocol]":
    if loader is not None:
        if file_ is not None and loader is EnvLoader:
            msg = (
                "EnvLoader reads from environment variables and does not use files. "
                "Remove file_ or use a file-based loader instead (e.g. EnvFileLoader)."
            )
            raise ValueError(msg)
        return loader

    if file_ is None:
        return EnvLoader

    file_path = Path(file_)

    if file_path.name.startswith(".env"):
        return EnvFileLoader

    return _resolve_by_extension(file_path.suffix.lower())



import configparser
from datetime import date, datetime, time
from pathlib import Path
from typing import cast

from adaptix import loader
from adaptix.provider import Provider

from dature.env_expand import expand_env_vars
from dature.path_finders.ini_ import TablePathFinder
from dature.sources_loader.base import BaseLoader
from dature.sources_loader.loaders import (
    bool_loader,
    bytearray_from_json_string,
    date_from_string,
    datetime_from_string,
    none_from_empty_string,
    optional_from_empty_string,
    time_from_string,
)
from dature.types import JSONValue


class IniLoader(BaseLoader):
    display_name = "ini"
    path_finder_class = TablePathFinder

    def _additional_loaders(self) -> list[Provider]:
        return [
            loader(date, date_from_string),
            loader(datetime, datetime_from_string),
            loader(time, time_from_string),
            loader(bytearray, bytearray_from_json_string),
            loader(type(None), none_from_empty_string),
            loader(str | None, optional_from_empty_string),
            loader(bool, bool_loader),
        ]

    def _pre_processing(self, data: JSONValue) -> JSONValue:
        prefixed = self._apply_prefix(data)
        expanded = expand_env_vars(prefixed, mode=self._expand_env_vars_mode)
        return self._parse_string_values(expanded)

    def _load(self, path: Path) -> JSONValue:
        config = configparser.ConfigParser(interpolation=None)
        with path.open() as f:
            config.read_file(f)
        if self._prefix and self._prefix in config:
            result: dict[str, JSONValue] = dict(config[self._prefix])
            child_prefix = self._prefix + "."
            for section in config.sections():
                if section.startswith(child_prefix):
                    nested_key = section[len(child_prefix) :]
                    result[nested_key] = dict(config[section])
            return {self._prefix: result}

        all_sections: dict[str, JSONValue] = {}
        if config.defaults():
            all_sections["DEFAULT"] = dict(config.defaults())
        for section in config.sections():
            parts = section.split(".")
            target = all_sections
            for part in parts[:-1]:
                if part not in target:
                    target[part] = {}
                target = cast("dict[str, JSONValue]", target[part])
            target[parts[-1]] = dict(config[section])
        return all_sections



import copy
from collections.abc import Callable, Sequence
from dataclasses import dataclass, replace
from typing import cast

from adaptix import Retort
from adaptix._internal.common import Loader
from adaptix._internal.model_tools.definitions import DefaultValue, InputShape, Param, ParamKind
from adaptix._internal.morphing.model.loader_provider import ModelLoaderProvider
from adaptix._internal.morphing.request_cls import LoaderRequest
from adaptix._internal.provider.essential import Mediator, Provider, RequestHandlerRegisterRecord
from adaptix._internal.provider.located_request import LocatedRequest
from adaptix._internal.provider.request_checkers import AlwaysTrueRequestChecker
from adaptix._internal.provider.shape_provider import InputShapeRequest, provide_generic_resolved_shape
from adaptix.load_error import LoadError

from dature.protocols import DataclassInstance
from dature.types import NOT_LOADED, JSONValue, NotLoaded, ProbeDict


class SkipFieldProvider(Provider):
    @staticmethod
    def _wrap_handler(
        mediator: Mediator[Loader[JSONValue | NotLoaded]],
        _request: LoaderRequest,
    ) -> Callable[[JSONValue], JSONValue | NotLoaded]:
        next_handler = mediator.provide_from_next()

        def chain_handler(data: JSONValue) -> JSONValue | NotLoaded:
            try:
                return cast("JSONValue", next_handler(data))
            except LoadError:
                return NOT_LOADED

        return chain_handler

    def get_request_handlers(self) -> Sequence[RequestHandlerRegisterRecord]:
        return [(LoaderRequest, AlwaysTrueRequestChecker(), self._wrap_handler)]


class ModelToDictProvider(ModelLoaderProvider):  # type: ignore[no-untyped-call]
    def _fetch_shape(
        self,
        mediator: Mediator[Loader[ProbeDict]],
        request: LocatedRequest[Loader[ProbeDict]],
    ) -> InputShape[ProbeDict]:
        shape = provide_generic_resolved_shape(
            mediator,
            InputShapeRequest(loc_stack=request.loc_stack),
        )
        optional_fields = tuple(
            replace(
                f,
                is_required=False,
                default=DefaultValue(NOT_LOADED),
            )
            for f in shape.fields
        )
        optional_params = tuple(Param(field_id=f.id, name=f.id, kind=ParamKind.KW_ONLY) for f in optional_fields)
        return replace(
            shape,
            fields=optional_fields,
            params=optional_params,
            constructor=dict,
            kwargs=None,
        )


def _collect_not_loaded_paths(data: ProbeDict, prefix: str) -> list[str]:
    paths: list[str] = []

    for key, value in data.items():
        current_path = f"{prefix}.{key}" if prefix else key
        if value is NOT_LOADED:
            paths.append(current_path)
        elif isinstance(value, dict):
            paths.extend(_collect_not_loaded_paths(value, current_path))

    return paths


def _remove_path_from_dict(data: dict[str, JSONValue], path: str) -> None:
    parts = path.split(".")
    current: dict[str, JSONValue] = data
    for part in parts[:-1]:
        next_val = current.get(part)
        if not isinstance(next_val, dict):
            return
        current = next_val

    current.pop(parts[-1], None)


@dataclass(frozen=True, slots=True)
class FilterResult:
    cleaned_dict: JSONValue
    skipped_paths: list[str]


def filter_invalid_fields(
    raw_dict: JSONValue,
    probe_retort: Retort,
    dataclass_: type[DataclassInstance],
    allowed_fields: set[str] | None,
) -> FilterResult:
    if not isinstance(raw_dict, dict):
        return FilterResult(cleaned_dict=raw_dict, skipped_paths=[])

    probed: ProbeDict = probe_retort.load(raw_dict, dataclass_)
    all_not_loaded = _collect_not_loaded_paths(probed, "")

    skipped: list[str] = []
    for path in all_not_loaded:
        if allowed_fields is not None and path not in allowed_fields:
            continue
        skipped.append(path)

    if not skipped:
        return FilterResult(cleaned_dict=raw_dict, skipped_paths=[])

    cleaned: dict[str, JSONValue] = copy.deepcopy(raw_dict)
    for path in skipped:
        _remove_path_from_dict(cleaned, path)

    return FilterResult(cleaned_dict=cleaned, skipped_paths=skipped)



import json
from datetime import date, datetime, time

# Expected number of time parts in HH:MM:SS format
TIME_PARTS_WITH_SECONDS = 3

# Expected number of time parts in HH:MM format
TIME_PARTS_WITHOUT_SECONDS = 2


def date_from_string(value: str) -> date:
    return date.fromisoformat(value)


def datetime_from_string(value: str) -> datetime:
    return datetime.fromisoformat(value)


def time_from_string(value: str) -> time:
    parts = value.split(":")
    if len(parts) == TIME_PARTS_WITH_SECONDS:
        return time(int(parts[0]), int(parts[1]), int(parts[2]))
    if len(parts) == TIME_PARTS_WITHOUT_SECONDS:
        return time(int(parts[0]), int(parts[1]))
    msg = f"Invalid time format: {value}"
    raise ValueError(msg)


def date_passthrough(value: date) -> date:
    return value


def datetime_passthrough(value: datetime) -> datetime:
    return value


def bytearray_from_string(value: str) -> bytearray:
    return bytearray(value, "utf-8")


def none_from_empty_string(value: str) -> None:
    if value == "":
        return
    msg = f"Cannot convert {value!r} to None"
    raise TypeError(msg)


def optional_from_empty_string(value: str) -> str | None:
    if value == "":
        return None
    return value


def _bool_from_string(value: str) -> bool:
    lower = value.lower().strip()
    if lower in ("true", "1", "yes", "on"):
        return True
    if lower in ("false", "0", "no", "off", ""):
        return False
    msg = f"Cannot convert {value!r} to bool"
    raise TypeError(msg)


def bool_loader(value: str | bool) -> bool:  # noqa: FBT001
    if isinstance(value, bool):
        return value
    return _bool_from_string(value)


def bytearray_from_json_string(value: str) -> bytearray:
    if value == "":
        return bytearray()

    if value.startswith("["):
        items = json.loads(value)
        if not isinstance(items, list):
            msg = f"Expected list in JSON, got {type(items)}"
            raise TypeError(msg)
        return bytearray(items)

    return bytearray(value.encode("utf-8"))



import abc

from dature.errors import LineRange
from dature.path_finders.base import PathFinder


class LinePathFinder(PathFinder):
    def __init__(self, content: str) -> None:
        self._lines = content.splitlines()

    def find_line_range(self, target_path: list[str]) -> LineRange | None:
        for i, line in enumerate(self._lines, 1):
            result = self._process_line(i, line, target_path)
            if result is not None:
                return result
        return None

    @abc.abstractmethod
    def _process_line(self, line_num: int, line: str, target_path: list[str]) -> LineRange | None: ...



from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent

import pytest
from adaptix import Retort
from adaptix.load_error import AggregateLoadError, LoadError

from dature import LoadMetadata, load
from dature.error_formatter import ErrorContext, extract_field_errors, resolve_source_location
from dature.errors import DatureConfigError, FieldLoadError, LineRange, SourceLocation
from dature.path_finders.json_ import JsonPathFinder
from dature.path_finders.toml_ import TomlPathFinder


class TestExtractFieldErrors:
    def test_type_error(self):
        @dataclass
        class Config:
            timeout: int

        r = Retort(strict_coercion=False)
        try:
            r.load({"timeout": "abc"}, Config)
        except (AggregateLoadError, LoadError) as exc:
            errors = extract_field_errors(exc)
            assert len(errors) == 1
            assert errors[0].field_path == ["timeout"]
            assert errors[0].message == "Bad string format"

    def test_missing_field(self):
        @dataclass
        class Config:
            timeout: int
            name: str

        r = Retort(strict_coercion=False)
        try:
            r.load({"timeout": 123}, Config)
        except (AggregateLoadError, LoadError) as exc:
            errors = extract_field_errors(exc)
            assert len(errors) == 1
            assert errors[0].field_path == ["name"]
            assert errors[0].message == "Missing required field"

    def test_nested_errors(self):
        @dataclass
        class DB:
            host: str
            port: int

        @dataclass
        class Config:
            timeout: int
            db: DB

        r = Retort(strict_coercion=False)
        try:
            r.load({"timeout": "abc", "db": {"host": "ok", "port": "xyz"}}, Config)
        except (AggregateLoadError, LoadError) as exc:
            errors = extract_field_errors(exc)
            assert len(errors) == 2
            paths = [e.field_path for e in errors]
            assert ["timeout"] in paths
            assert ["db", "port"] in paths

    def test_multiple_missing_fields(self):
        @dataclass
        class Config:
            a: int
            b: str
            c: float

        r = Retort(strict_coercion=False)
        try:
            r.load({}, Config)
        except (AggregateLoadError, LoadError) as exc:
            errors = extract_field_errors(exc)
            assert len(errors) == 3
            paths = sorted([e.field_path[0] for e in errors])
            assert paths == ["a", "b", "c"]


class TestResolveSourceLocation:
    def test_env_source(self):
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="env",
            file_path=None,
            prefix="APP_",
            split_symbols="__",
            path_finder_class=None,
        )
        loc = resolve_source_location(["database", "port"], ctx, file_content=None)
        assert loc.source_type == "env"
        assert loc.env_var_name == "APP_DATABASE__PORT"
        assert loc.file_path is None

    def test_env_source_no_prefix(self):
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="env",
            file_path=None,
            prefix=None,
            split_symbols="__",
            path_finder_class=None,
        )
        loc = resolve_source_location(["timeout"], ctx, file_content=None)
        assert loc.env_var_name == "TIMEOUT"

    def test_env_source_custom_split_symbols(self):
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="env",
            file_path=None,
            prefix="APP_",
            split_symbols="_",
            path_finder_class=None,
        )
        loc = resolve_source_location(["database", "port"], ctx, file_content=None)
        assert loc.env_var_name == "APP_DATABASE_PORT"

    def test_json_source_with_line(self):
        content = '{\n  "timeout": "30",\n  "name": "test"\n}'
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="json",
            file_path=Path("config.json"),
            prefix=None,
            split_symbols="__",
            path_finder_class=JsonPathFinder,
        )
        loc = resolve_source_location(["timeout"], ctx, file_content=content)
        assert loc.source_type == "json"
        assert loc.line_range == LineRange(start=2, end=2)
        assert loc.line_content == ['"timeout": "30",']

    def test_toml_source_with_line(self):
        content = 'timeout = "30"\nname = "test"'
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="toml",
            file_path=Path("config.toml"),
            prefix=None,
            split_symbols="__",
            path_finder_class=TomlPathFinder,
        )
        loc = resolve_source_location(["timeout"], ctx, file_content=content)
        assert loc.source_type == "toml"
        assert loc.line_range == LineRange(start=1, end=1)
        assert loc.line_content == ['timeout = "30"']

    def test_envfile_source(self):
        content = "# comment\nAPP_TIMEOUT=30\nAPP_NAME=test"
        ctx = ErrorContext(
            dataclass_name="Config",
            loader_type="envfile",
            file_path=Path(".env"),
            prefix="APP_",
            split_symbols="__",
            path_finder_class=None,
        )
        loc = resolve_source_location(["timeout"], ctx, file_content=content)
        assert loc.source_type == "envfile"
        assert loc.env_var_name == "APP_TIMEOUT"
        assert loc.line_range == LineRange(start=2, end=2)
        assert loc.line_content == ["APP_TIMEOUT=30"]


class TestDatureConfigErrorFormat:
    def test_single_error_message(self):
        errors = [
            FieldLoadError(
                field_path=["timeout"],
                message="Expected int, got str",
                input_value="30",
                location=SourceLocation(
                    source_type="toml",
                    file_path=Path("config.toml"),
                    line_range=LineRange(start=2, end=2),
                    line_content=['timeout = "30"'],
                    env_var_name=None,
                ),
            ),
        ]
        exc = DatureConfigError("Config", errors)
        assert str(exc) == dedent("""\
            Config loading errors (1)

              [timeout]  Expected int, got str
               └── FILE 'config.toml', line 2
                   timeout = "30"
            """)

    def test_multiple_errors_message(self):
        errors = [
            FieldLoadError(
                field_path=["timeout"],
                message="Bad string format",
                input_value="abc",
                location=SourceLocation(
                    source_type="json",
                    file_path=Path("config.json"),
                    line_range=LineRange(start=2, end=2),
                    line_content=['"timeout": "abc"'],
                    env_var_name=None,
                ),
            ),
            FieldLoadError(
                field_path=["db", "port"],
                message="Missing required field",
                input_value=None,
                location=SourceLocation(
                    source_type="json",
                    file_path=Path("config.json"),
                    line_range=None,
                    line_content=None,
                    env_var_name=None,
                ),
            ),
        ]
        exc = DatureConfigError("Config", errors)
        assert str(exc) == dedent("""\
            Config loading errors (2)

              [timeout]  Bad string format
               └── FILE 'config.json', line 2
                   "timeout": "abc"

              [db.port]  Missing required field
               └── FILE 'config.json'
            """)

    def test_env_error_message(self):
        errors = [
            FieldLoadError(
                field_path=["database", "port"],
                message="Bad string format",
                input_value="abc",
                location=SourceLocation(
                    source_type="env",
                    file_path=None,
                    line_range=None,
                    line_content=None,
                    env_var_name="APP_DATABASE__PORT",
                ),
            ),
        ]
        exc = DatureConfigError("Config", errors)
        assert str(exc) == dedent("""\
            Config loading errors (1)

              [database.port]  Bad string format
               └── ENV 'APP_DATABASE__PORT'
            """)


class TestLoadIntegrationErrors:
    def test_json_type_error_decorator(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"timeout": "abc", "name": "test"}')

        metadata = LoadMetadata(file_=str(json_file))

        @load(metadata)
        @dataclass
        class Config:
            timeout: int
            name: str

        with pytest.raises(DatureConfigError) as exc_info:
            Config()

        err = exc_info.value
        assert len(err.exceptions) == 1
        first = err.exceptions[0]
        assert isinstance(first, FieldLoadError)
        assert first.field_path == ["timeout"]
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [timeout]  Bad string format
               └── FILE '{json_file}', line 1
                   {json_file.read_text()}
            """)

    def test_json_missing_field_function(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test"}')

        @dataclass
        class Config:
            name: str
            port: int

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert len(err.exceptions) == 1
        first = err.exceptions[0]
        assert isinstance(first, FieldLoadError)
        assert first.field_path == ["port"]
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [port]  Missing required field
               └── FILE '{json_file}'
            """)

    def test_multiple_errors_at_once(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"timeout": "abc"}')

        @dataclass
        class Config:
            timeout: int
            name: str

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert len(err.exceptions) == 2
        paths = [e.field_path for e in err.exceptions if isinstance(e, FieldLoadError)]
        assert ["timeout"] in paths
        assert ["name"] in paths
        assert str(err) == dedent(f"""\
            Config loading errors (2)

              [timeout]  Bad string format
               └── FILE '{json_file}', line 1
                   {json_file.read_text()}

              [name]  Missing required field
               └── FILE '{json_file}'
            """)

    def test_nested_dataclass_error(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{\n  "db": {\n    "host": "localhost",\n    "port": "abc"\n  }\n}')

        @dataclass
        class DB:
            host: str
            port: int

        @dataclass
        class Config:
            db: DB

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert len(err.exceptions) == 1
        first = err.exceptions[0]
        assert isinstance(first, FieldLoadError)
        assert first.field_path == ["db", "port"]
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [db.port]  Bad string format
               └── FILE '{json_file}', line 4
                   "port": "abc"
            """)

    def test_env_type_error(self, monkeypatch: pytest.MonkeyPatch):
        monkeypatch.setenv("APP_TIMEOUT", "abc")
        monkeypatch.setenv("APP_NAME", "test")

        metadata = LoadMetadata(prefix="APP_")

        @load(metadata)
        @dataclass
        class Config:
            timeout: int
            name: str

        with pytest.raises(DatureConfigError) as exc_info:
            Config()

        err = exc_info.value
        assert len(err.exceptions) == 1
        assert str(err) == dedent("""\
            Config loading errors (1)

              [timeout]  Bad string format
               └── ENV 'APP_TIMEOUT'
            """)

    def test_toml_with_line_number(self, tmp_path: Path):
        toml_file = tmp_path / "config.toml"
        toml_file.write_text('name = "test"\ntimeout = "abc"\n')

        @dataclass
        class Config:
            name: str
            timeout: int

        metadata = LoadMetadata(file_=str(toml_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert len(err.exceptions) == 1
        first = err.exceptions[0]
        assert isinstance(first, FieldLoadError)
        assert first.location is not None
        assert first.location.line_range == LineRange(start=2, end=2)
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [timeout]  Bad string format
               └── FILE '{toml_file}', line 2
                   timeout = "abc"
            """)

    def test_json_with_line_number(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{\n  "name": "test",\n  "timeout": "abc"\n}')

        @dataclass
        class Config:
            name: str
            timeout: int

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        first = err.exceptions[0]
        assert isinstance(first, FieldLoadError)
        assert first.location is not None
        assert first.location.line_range == LineRange(start=3, end=3)
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [timeout]  Bad string format
               └── FILE '{json_file}', line 3
                   "timeout": "abc"
            """)


class TestLineTruncation:
    @pytest.mark.parametrize(
        ("line_content", "expected_content"),
        [
            pytest.param(
                "a" * 80,
                "a" * 80,
                id="exactly_80_chars_not_truncated",
            ),
            pytest.param(
                "b" * 81,
                "b" * 77 + "...",
                id="81_chars_truncated",
            ),
            pytest.param(
                "c" * 120,
                "c" * 77 + "...",
                id="120_chars_truncated",
            ),
        ],
    )
    def test_file_source_truncation(
        self,
        line_content: str,
        expected_content: str,
    ) -> None:
        errors = [
            FieldLoadError(
                field_path=["timeout"],
                message="Expected int, got str",
                input_value="30",
                location=SourceLocation(
                    source_type="toml",
                    file_path=Path("config.toml"),
                    line_range=LineRange(start=2, end=2),
                    line_content=[line_content],
                    env_var_name=None,
                ),
            ),
        ]
        exc = DatureConfigError("Config", errors)
        assert str(exc) == dedent(f"""\
            Config loading errors (1)

              [timeout]  Expected int, got str
               └── FILE 'config.toml', line 2
                   {expected_content}
            """)

    @pytest.mark.parametrize(
        ("line_content", "expected_content"),
        [
            pytest.param(
                "a" * 80,
                "a" * 80,
                id="exactly_80_chars_not_truncated",
            ),
            pytest.param(
                "b" * 81,
                "b" * 77 + "...",
                id="81_chars_truncated",
            ),
            pytest.param(
                "c" * 120,
                "c" * 77 + "...",
                id="120_chars_truncated",
            ),
        ],
    )
    def test_envfile_source_truncation(
        self,
        line_content: str,
        expected_content: str,
    ) -> None:
        errors = [
            FieldLoadError(
                field_path=["timeout"],
                message="Bad string format",
                input_value="abc",
                location=SourceLocation(
                    source_type="envfile",
                    file_path=Path(".env"),
                    line_range=LineRange(start=2, end=2),
                    line_content=[line_content],
                    env_var_name="APP_TIMEOUT",
                ),
            ),
        ]
        exc = DatureConfigError("Config", errors)
        assert str(exc) == dedent(f"""\
            Config loading errors (1)

              [timeout]  Bad string format
               └── ENV FILE '.env', var 'APP_TIMEOUT'
                   {expected_content}
            """)

    def test_multiline_content_each_line_truncated(self) -> None:
        line_short = "short line"
        line_long = "x" * 100
        errors = [
            FieldLoadError(
                field_path=["db"],
                message="Expected int, got dict",
                input_value=None,
                location=SourceLocation(
                    source_type="json",
                    file_path=Path("config.json"),
                    line_range=LineRange(start=2, end=4),
                    line_content=[line_long, line_short, line_long],
                    env_var_name=None,
                ),
            ),
        ]
        exc = DatureConfigError("Config", errors)
        truncated = "x" * 77 + "..."
        assert str(exc) == dedent(f"""\
            Config loading errors (1)

              [db]  Expected int, got dict
               └── FILE 'config.json', line 2-4
                   {truncated}
                   {line_short}
                   {truncated}
            """)

    def test_four_lines_shows_two_and_ellipsis(self) -> None:
        errors = [
            FieldLoadError(
                field_path=["db"],
                message="Expected int, got dict",
                input_value=None,
                location=SourceLocation(
                    source_type="json",
                    file_path=Path("config.json"),
                    line_range=LineRange(start=2, end=5),
                    line_content=["line1", "line2", "line3", "line4"],
                    env_var_name=None,
                ),
            ),
        ]
        exc = DatureConfigError("Config", errors)
        assert str(exc) == dedent("""\
            Config loading errors (1)

              [db]  Expected int, got dict
               └── FILE 'config.json', line 2-5
                   line1
                   line2
                   ...
            """)

    def test_five_lines_shows_two_and_ellipsis(self) -> None:
        errors = [
            FieldLoadError(
                field_path=["db"],
                message="Expected int, got dict",
                input_value=None,
                location=SourceLocation(
                    source_type="json",
                    file_path=Path("config.json"),
                    line_range=LineRange(start=2, end=6),
                    line_content=["line1", "line2", "line3", "line4", "line5"],
                    env_var_name=None,
                ),
            ),
        ]
        exc = DatureConfigError("Config", errors)
        assert str(exc) == dedent("""\
            Config loading errors (1)

              [db]  Expected int, got dict
               └── FILE 'config.json', line 2-6
                   line1
                   line2
                   ...
            """)


class TestMultilineValueDisplay:
    def test_json_multiline_dict(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{\n  "db": {\n    "host": "localhost",\n    "port": "abc"\n  }\n}')

        @dataclass
        class Config:
            db: int

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [db]  Expected int | float | str, got dict
               └── FILE '{json_file}', line 2-5
                   "db": {{
                     "host": "localhost",
                   ...
            """)

    def test_yaml_multiline_block(self, tmp_path: Path):
        yaml_file = tmp_path / "config.yaml"
        yaml_file.write_text("db:\n  host: localhost\n  port: abc\nname: test\n")

        @dataclass
        class Config:
            db: int
            name: str

        metadata = LoadMetadata(file_=str(yaml_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [db]  Expected int | float | str, got dict
               └── FILE '{yaml_file}', line 1-3
                   db:
                     host: localhost
                     port: abc
            """)

    def test_toml_multiline_array(self, tmp_path: Path):
        toml_file = tmp_path / "config.toml"
        toml_file.write_text('tags = [\n  "a",\n  "b"\n]\n')

        @dataclass
        class Config:
            tags: int

        metadata = LoadMetadata(file_=str(toml_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [tags]  Expected int | float | str, got list
               └── FILE '{toml_file}', line 1-4
                   tags = [
                     "a",
                   ...
            """)

    def test_json_multiline_array(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{\n  "tags": [\n    "a",\n    "b"\n  ]\n}')

        @dataclass
        class Config:
            tags: int

        metadata = LoadMetadata(file_=str(json_file))

        with pytest.raises(DatureConfigError) as exc_info:
            load(metadata, Config)

        err = exc_info.value
        assert str(err) == dedent(f"""\
            Config loading errors (1)

              [tags]  Expected int | float | str, got list
               └── FILE '{json_file}', line 2-5
                   "tags": [
                     "a",
                   ...
            """)



from collections.abc import Callable
from pathlib import Path
from typing import Any, overload

from dature.merge import merge_load_as_function, merge_make_decorator
from dature.metadata import LoadMetadata, MergeMetadata
from dature.patcher import load_as_function, make_decorator
from dature.protocols import DataclassInstance
from dature.sources_loader.resolver import resolve_loader


@overload
def load[T](
    metadata: LoadMetadata | MergeMetadata | tuple[LoadMetadata, ...] | None,
    /,
    dataclass_: type[T],
    *,
    debug: bool = False,
) -> T: ...


@overload
def load(
    metadata: LoadMetadata | MergeMetadata | tuple[LoadMetadata, ...] | None = None,
    /,
    dataclass_: None = None,
    *,
    cache: bool = True,
    debug: bool = False,
) -> Callable[[type[DataclassInstance]], type[DataclassInstance]]: ...


def load(
    metadata: LoadMetadata | MergeMetadata | tuple[LoadMetadata, ...] | None = None,
    /,
    dataclass_: type[Any] | None = None,
    *,
    cache: bool = True,
    debug: bool = False,
) -> Any:
    if isinstance(metadata, tuple):
        metadata = MergeMetadata(sources=metadata)

    if isinstance(metadata, MergeMetadata):
        if dataclass_ is not None:
            return merge_load_as_function(metadata, dataclass_, debug=debug)
        return merge_make_decorator(metadata, cache=cache, debug=debug)

    if metadata is None:
        metadata = LoadMetadata()

    loader_instance = resolve_loader(metadata)
    file_path = Path(metadata.file_) if metadata.file_ else Path()

    if dataclass_ is not None:
        return load_as_function(
            loader_instance=loader_instance,
            file_path=file_path,
            dataclass_=dataclass_,
            metadata=metadata,
            debug=debug,
        )

    return make_decorator(
        loader_instance=loader_instance,
        file_path=file_path,
        metadata=metadata,
        cache=cache,
        debug=debug,
    )



from dataclasses import dataclass

from dature.errors import FieldGroupError, FieldGroupViolationError
from dature.predicate import ResolvedFieldGroup
from dature.types import JSONValue

_SENTINEL = object()


@dataclass(frozen=True, slots=True)
class FieldGroupContext:
    source_reprs: tuple[str, ...]
    field_origins: dict[str, int]
    dataclass_name: str


def _get_nested_value(data: JSONValue, dot_path: str) -> object:
    if not isinstance(data, dict):
        return _SENTINEL
    parts = dot_path.split(".")
    current: JSONValue = data
    for part in parts:
        if not isinstance(current, dict) or part not in current:
            return _SENTINEL
        current = current[part]
    return current


def validate_field_groups(
    *,
    base: JSONValue,
    source: JSONValue,
    field_group_paths: tuple[ResolvedFieldGroup, ...],
    source_index: int,
    ctx: FieldGroupContext,
) -> None:
    violations: list[FieldGroupViolationError] = []
    current_source_repr = ctx.source_reprs[source_index]

    for group in field_group_paths:
        changed: list[str] = []
        changed_sources: list[str] = []
        unchanged: list[str] = []
        unchanged_sources: list[str] = []

        for path in group.paths:
            source_val = _get_nested_value(source, path)
            if source_val is _SENTINEL:
                unchanged.append(path)
                origin_idx = ctx.field_origins.get(path)
                if origin_idx is not None:
                    unchanged_sources.append(ctx.source_reprs[origin_idx])
                else:
                    unchanged_sources.append("none")
                continue
            base_val = _get_nested_value(base, path)
            if source_val == base_val:
                unchanged.append(path)
                origin_idx = ctx.field_origins.get(path)
                if origin_idx is not None:
                    unchanged_sources.append(ctx.source_reprs[origin_idx])
                else:
                    unchanged_sources.append(current_source_repr)
            else:
                changed.append(path)
                changed_sources.append(current_source_repr)

        if changed and unchanged:
            violations.append(
                FieldGroupViolationError(
                    group_fields=group.paths,
                    changed_fields=tuple(changed),
                    unchanged_fields=tuple(unchanged),
                    changed_sources=tuple(changed_sources),
                    unchanged_sources=tuple(unchanged_sources),
                    source_index=source_index,
                ),
            )

    if violations:
        raise FieldGroupError(ctx.dataclass_name, violations)



from typing import Any

from dature.loader_resolver import resolve_loader_class
from dature.metadata import LoadMetadata
from dature.protocols import LoaderProtocol
from dature.sources_loader.env_ import EnvLoader
from dature.types import ExpandEnvVarsMode


def resolve_loader(
    metadata: LoadMetadata,
    *,
    expand_env_vars: ExpandEnvVarsMode | None = None,
) -> LoaderProtocol:
    loader_class = resolve_loader_class(metadata.loader, metadata.file_)

    resolved_expand = expand_env_vars or metadata.expand_env_vars or "default"

    kwargs: dict[str, Any] = {
        "prefix": metadata.prefix,
        "name_style": metadata.name_style,
        "field_mapping": metadata.field_mapping,
        "root_validators": metadata.root_validators,
        "validators": metadata.validators,
        "expand_env_vars": resolved_expand,
    }

    if issubclass(loader_class, EnvLoader):
        kwargs["split_symbols"] = metadata.split_symbols

    return loader_class(**kwargs)



from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True, slots=True, kw_only=True)
class MinItems:
    value: int
    error_message: str = "Value must have at least {value} items"

    def get_validator_func(self) -> Callable[[Sequence[Any]], bool]:
        def validate(val: Sequence[Any]) -> bool:
            return len(val) >= self.value

        return validate

    def get_error_message(self) -> str:
        return self.error_message.format(value=self.value)


@dataclass(frozen=True, slots=True, kw_only=True)
class MaxItems:
    value: int
    error_message: str = "Value must have at most {value} items"

    def get_validator_func(self) -> Callable[[Sequence[Any]], bool]:
        def validate(val: Sequence[Any]) -> bool:
            return len(val) <= self.value

        return validate

    def get_error_message(self) -> str:
        return self.error_message.format(value=self.value)


@dataclass(frozen=True, slots=True, kw_only=True)
class UniqueItems:
    error_message: str = "Value must contain unique items"

    def get_validator_func(self) -> Callable[[Sequence[Any]], bool]:
        def validate(val: Sequence[Any]) -> bool:
            return len(val) == len(set(val))

        return validate

    def get_error_message(self) -> str:
        return self.error_message



import re
from typing import ClassVar


class SecretStr:
    __slots__ = ("_secret_value",)

    def __init__(self, secret_value: str) -> None:
        self._secret_value = secret_value

    def get_secret_value(self) -> str:
        return self._secret_value

    def __str__(self) -> str:
        return "**********"

    def __repr__(self) -> str:
        return "SecretStr('**********')"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SecretStr):
            return NotImplemented
        return self._secret_value == other._secret_value

    def __hash__(self) -> int:
        return hash(self._secret_value)

    def __len__(self) -> int:
        return len(self._secret_value)


class PaymentCardNumber:
    __slots__ = ("_number",)

    _MIN_LENGTH: ClassVar[int] = 12
    _MAX_LENGTH: ClassVar[int] = 19
    _LUHN_DOUBLE_THRESHOLD: ClassVar[int] = 9

    _BRAND_RULES: ClassVar[list[tuple[str, int, int, int]]] = [
        # 6 digits
        ("Verve", 6, 506099, 506198),
        ("Verve", 6, 650002, 650027),
        ("Discover", 6, 622126, 622925),
        # 4 digits
        ("Mir", 4, 2200, 2204),
        ("JCB", 4, 3528, 3589),
        ("Mastercard", 4, 2221, 2720),
        ("Discover", 4, 6011, 6011),
        ("Maestro", 4, 5018, 5018),
        ("Maestro", 4, 5020, 5020),
        ("Maestro", 4, 5038, 5038),
        ("Maestro", 4, 5893, 5893),
        ("Maestro", 4, 6304, 6304),
        ("Maestro", 4, 6759, 6759),
        ("Maestro", 4, 6761, 6763),
        ("Troy", 4, 9792, 9792),
        # 3 digits
        ("Diners Club", 3, 300, 305),
        ("Discover", 3, 644, 649),
        # 2 digits
        ("American Express", 2, 34, 34),
        ("American Express", 2, 37, 37),
        ("Mastercard", 2, 51, 55),
        ("RuPay", 2, 60, 60),
        ("UnionPay", 2, 62, 62),
        ("Discover", 2, 65, 65),
        ("Diners Club", 2, 36, 36),
        ("Diners Club", 2, 38, 38),
        ("Maestro", 2, 67, 67),
        # 1 digit
        ("Visa", 1, 4, 4),
    ]

    def __init__(self, card_number: str) -> None:
        cleaned = card_number.replace(" ", "").replace("-", "")

        if not cleaned.isdigit():
            msg = f"Card number must contain only digits, got: {card_number!r}"
            raise ValueError(msg)

        if not (self._MIN_LENGTH <= len(cleaned) <= self._MAX_LENGTH):
            msg = f"Card number must be {self._MIN_LENGTH}-{self._MAX_LENGTH} digits, got {len(cleaned)}"
            raise ValueError(msg)

        if not self._luhn_check(cleaned):
            msg = f"Card number failed Luhn check: {card_number!r}"
            raise ValueError(msg)

        self._number = cleaned

    @staticmethod
    def _luhn_check(number: str) -> bool:
        total = 0
        for i, digit_char in enumerate(reversed(number)):
            digit = int(digit_char)
            if i % 2 == 1:
                digit *= 2
                if digit > PaymentCardNumber._LUHN_DOUBLE_THRESHOLD:
                    digit -= PaymentCardNumber._LUHN_DOUBLE_THRESHOLD
            total += digit
        return total % 10 == 0

    @property
    def masked(self) -> str:
        return "*" * (len(self._number) - 4) + self._number[-4:]

    @property
    def brand(self) -> str:
        for brand_name, prefix_len, prefix_min, prefix_max in self._BRAND_RULES:
            if len(self._number) < prefix_len:
                continue
            prefix = int(self._number[:prefix_len])
            if prefix_min <= prefix <= prefix_max:
                return brand_name
        return "Unknown"

    def __str__(self) -> str:
        return self.masked

    def __repr__(self) -> str:
        return f"PaymentCardNumber('{self.masked}')"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, PaymentCardNumber):
            return NotImplemented
        return self._number == other._number

    def __hash__(self) -> int:
        return hash(self._number)


class ByteSize:
    __slots__ = ("_bytes",)

    _UNITS: ClassVar[dict[str, int]] = {
        "b": 1,
        "kb": 10**3,
        "mb": 10**6,
        "gb": 10**9,
        "tb": 10**12,
        "pb": 10**15,
        "kib": 2**10,
        "mib": 2**20,
        "gib": 2**30,
        "tib": 2**40,
        "pib": 2**50,
    }

    _PARSE_RE: ClassVar[re.Pattern[str]] = re.compile(
        r"^\s*([0-9]*\.?[0-9]+)\s*([a-zA-Z]*)\s*$",
    )

    def __init__(self, value: int | str) -> None:
        if isinstance(value, int):
            self._bytes = value
        else:
            self._bytes = self._parse(value)

    @staticmethod
    def _parse(value: str) -> int:
        match = ByteSize._PARSE_RE.match(value)
        if match is None:
            msg = f"Invalid byte size format: {value!r}"
            raise ValueError(msg)

        number_str = match.group(1)
        unit_str = match.group(2).lower()

        if unit_str == "":
            unit_str = "b"

        if unit_str not in ByteSize._UNITS:
            msg = f"Unknown unit: {unit_str!r}"
            raise ValueError(msg)

        multiplier = ByteSize._UNITS[unit_str]
        return int(float(number_str) * multiplier)

    def human_readable(self, *, decimal: bool = False) -> str:
        if decimal:
            units = [("PB", 10**15), ("TB", 10**12), ("GB", 10**9), ("MB", 10**6), ("KB", 10**3)]
        else:
            units = [("PiB", 2**50), ("TiB", 2**40), ("GiB", 2**30), ("MiB", 2**20), ("KiB", 2**10)]

        for unit_name, threshold in units:
            if self._bytes >= threshold:
                size = self._bytes / threshold
                if size == int(size):
                    return f"{int(size)}{unit_name}"
                return f"{size:.1f}{unit_name}"

        return f"{self._bytes}B"

    def __int__(self) -> int:
        return self._bytes

    def __str__(self) -> str:
        return self.human_readable()

    def __repr__(self) -> str:
        return f"ByteSize({self._bytes})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ByteSize):
            return NotImplemented
        return self._bytes == other._bytes

    def __hash__(self) -> int:
        return hash(self._bytes)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, ByteSize):
            return NotImplemented
        return self._bytes < other._bytes

    def __le__(self, other: object) -> bool:
        if not isinstance(other, ByteSize):
            return NotImplemented
        return self._bytes <= other._bytes

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, ByteSize):
            return NotImplemented
        return self._bytes > other._bytes

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, ByteSize):
            return NotImplemented
        return self._bytes >= other._bytes



from collections.abc import Callable
from dataclasses import dataclass


@dataclass(frozen=True, slots=True, kw_only=True)
class Gt:
    value: int | float
    error_message: str = "Value must be greater than {value}"

    def get_validator_func(self) -> Callable[[int | float], bool]:
        def validate(val: float) -> bool:
            return val > self.value

        return validate

    def get_error_message(self) -> str:
        return self.error_message.format(value=self.value)


@dataclass(frozen=True, slots=True, kw_only=True)
class Ge:
    value: int | float
    error_message: str = "Value must be greater than or equal to {value}"

    def get_validator_func(self) -> Callable[[int | float], bool]:
        def validate(val: float) -> bool:
            return val >= self.value

        return validate

    def get_error_message(self) -> str:
        return self.error_message.format(value=self.value)


@dataclass(frozen=True, slots=True, kw_only=True)
class Lt:
    value: int | float
    error_message: str = "Value must be less than {value}"

    def get_validator_func(self) -> Callable[[int | float], bool]:
        def validate(val: float) -> bool:
            return val < self.value

        return validate

    def get_error_message(self) -> str:
        return self.error_message.format(value=self.value)


@dataclass(frozen=True, slots=True, kw_only=True)
class Le:
    value: int | float
    error_message: str = "Value must be less than or equal to {value}"

    def get_validator_func(self) -> Callable[[int | float], bool]:
        def validate(val: float) -> bool:
            return val <= self.value

        return validate

    def get_error_message(self) -> str:
        return self.error_message.format(value=self.value)



"""Tests for patcher.py."""

from dataclasses import dataclass
from pathlib import Path

import pytest

from dature.metadata import LoadMetadata
from dature.patcher import load_as_function, make_decorator
from dature.sources_loader.json_ import JsonLoader


class TestMakeDecorator:
    def test_not_dataclass_raises(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test"}')
        metadata = LoadMetadata(file_=str(json_file))

        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata)

        with pytest.raises(TypeError, match="must be a dataclass"):

            @decorator
            class NotADataclass:
                pass

    def test_patches_init(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test"}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str

        original_init = Config.__init__
        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata)
        decorator(Config)

        assert Config.__init__ is not original_init

    def test_patches_post_init(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test"}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str

        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata)
        decorator(Config)

        assert hasattr(Config, "__post_init__")

    def test_loads_on_init(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "from_file", "port": 8080}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str
            port: int

        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata)
        decorator(Config)

        config = Config()
        assert config.name == "from_file"
        assert config.port == 8080

    def test_init_args_override_loaded(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "from_file", "port": 8080}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str
            port: int

        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata)
        decorator(Config)

        config = Config(name="overridden")
        assert config.name == "overridden"
        assert config.port == 8080

    def test_returns_same_class(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test"}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str

        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata)
        result = decorator(Config)

        assert result is Config

    def test_preserves_original_post_init(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test"}')
        metadata = LoadMetadata(file_=str(json_file))

        post_init_called = []

        @dataclass
        class Config:
            name: str

            def __post_init__(self):
                post_init_called.append(True)

        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata)
        decorator(Config)

        Config()
        assert len(post_init_called) == 1


class TestCache:
    def test_cache_returns_same_data(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "original", "port": 8080}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str
            port: int

        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata, cache=True)
        decorator(Config)

        first = Config()
        json_file.write_text('{"name": "updated", "port": 9090}')
        second = Config()

        assert first.name == "original"
        assert second.name == "original"
        assert second.port == 8080

    def test_no_cache_rereads_file(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "original", "port": 8080}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str
            port: int

        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata, cache=False)
        decorator(Config)

        first = Config()
        json_file.write_text('{"name": "updated", "port": 9090}')
        second = Config()

        assert first.name == "original"
        assert second.name == "updated"
        assert second.port == 9090

    def test_cache_allows_override(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "original", "port": 8080}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str
            port: int

        decorator = make_decorator(loader_instance=JsonLoader(), file_path=json_file, metadata=metadata, cache=True)
        decorator(Config)

        first = Config()
        assert first.name == "original"
        assert first.port == 8080

        second = Config(name="overridden")
        assert second.name == "overridden"
        assert second.port == 8080


class TestLoadAsFunction:
    def test_returns_loaded_dataclass(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"name": "test", "port": 3000}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str
            port: int

        result = load_as_function(
            loader_instance=JsonLoader(),
            file_path=json_file,
            dataclass_=Config,
            metadata=metadata,
            debug=False,
        )

        assert result.name == "test"
        assert result.port == 3000

    def test_with_prefix(self, tmp_path: Path):
        json_file = tmp_path / "config.json"
        json_file.write_text('{"app": {"name": "nested"}}')
        metadata = LoadMetadata(file_=str(json_file))

        @dataclass
        class Config:
            name: str

        result = load_as_function(
            loader_instance=JsonLoader(prefix="app"),
            file_path=json_file,
            dataclass_=Config,
            metadata=metadata,
            debug=False,
        )

        assert result.name == "nested"



from dataclasses import dataclass, fields, is_dataclass
from typing import TYPE_CHECKING, get_type_hints

from dature.field_path import FieldPath, _resolve_field_type, validate_field_path_owner
from dature.protocols import DataclassInstance

if TYPE_CHECKING:
    from dature.metadata import FieldGroup, FieldMergeStrategy, MergeRule


@dataclass(frozen=True, slots=True)
class ResolvedFieldGroup:
    paths: tuple[str, ...]


def extract_field_path(predicate: object, dataclass_: type[DataclassInstance] | None = None) -> str:
    if not isinstance(predicate, FieldPath):
        msg = f"Expected FieldPath, got {type(predicate).__name__}"
        raise TypeError(msg)
    if dataclass_ is not None:
        validate_field_path_owner(predicate, dataclass_)
    return predicate.as_path()


def build_field_merge_map(
    field_merges: "tuple[MergeRule, ...]",
    dataclass_: type[DataclassInstance] | None = None,
) -> "dict[str, FieldMergeStrategy]":
    result: dict[str, FieldMergeStrategy] = {}
    for rule in field_merges:
        path = extract_field_path(rule.predicate, dataclass_)
        result[path] = rule.strategy
    return result


def _expand_dataclass_fields(prefix: str, dc_type: type) -> list[str]:
    result: list[str] = []
    hints = get_type_hints(dc_type)
    for f in fields(dc_type):
        child_path = f"{prefix}.{f.name}" if prefix else f.name
        child_type = hints.get(f.name)
        if child_type is not None and is_dataclass(child_type):
            result.extend(_expand_dataclass_fields(child_path, child_type))
        else:
            result.append(child_path)
    return result


def build_field_group_paths(
    field_groups: "tuple[FieldGroup, ...]",
    dataclass_: type[DataclassInstance],
) -> tuple[ResolvedFieldGroup, ...]:
    resolved: list[ResolvedFieldGroup] = []
    for group in field_groups:
        paths: list[str] = []
        for field in group.fields:
            path = extract_field_path(field, dataclass_)
            if isinstance(field, FieldPath) and isinstance(field.owner, type):
                resolved_type = _resolve_field_type(field.owner, field.parts)
            else:
                resolved_type = _resolve_field_type(dataclass_, tuple(path.split(".")))
            if resolved_type is not None:
                paths.extend(_expand_dataclass_fields(path, resolved_type))
            else:
                paths.append(path)
        resolved.append(ResolvedFieldGroup(paths=tuple(paths)))
    return tuple(resolved)



from dature.path_finders.char_base import (
    CharPathFinder,
    KeyParseResult,
    PosLine,
    StackEntry,
    build_path,
    increment_parent_array,
)


class JsonPathFinder(CharPathFinder):
    def _skip_noise(self, content: str, pos: int, line: int, length: int) -> PosLine | None:
        if pos >= length:
            return None
        ch = content[pos]
        if ch in " \t\r\n":
            if ch == "\n":
                line += 1
            return PosLine(pos=pos + 1, line=line)
        if ch == ",":
            return PosLine(pos=pos + 1, line=line)
        return None

    def _try_parse_key(
        self,
        content: str,
        pos: int,
        line: int,
        length: int,
        stack: list[StackEntry],
        target_path: list[str],
    ) -> KeyParseResult | None:
        if content[pos] != '"':
            return None

        string_start = pos + 1
        end_pos = _skip_string(content, string_start)
        string_value = content[string_start : end_pos - 1]

        colon_pos = _skip_ws(content, end_pos, length)

        if colon_pos >= length or content[colon_pos] != ":":
            increment_parent_array(stack)
            return KeyParseResult(pos=end_pos, line=line, last_key=None, found=False)

        if build_path(stack, string_value) == target_path:
            return KeyParseResult(pos=end_pos, line=line, last_key=None, found=True)

        return KeyParseResult(pos=colon_pos + 1, line=line, last_key=string_value, found=False)

    def _find_value_end_line(self, content: str, pos: int, length: int, start_line: int) -> int:
        pos, current_line = _skip_to_value_start(content, pos, length, start_line)
        if pos >= length:
            return current_line

        ch = content[pos]

        if ch == '"':
            return current_line

        if ch in "{[":
            return _scan_container_end(content, pos, length, current_line)

        return current_line


def _skip_string(content: str, pos: int) -> int:
    """Advance past a JSON string body. Returns position after the closing quote."""
    length = len(content)
    while pos < length:
        ch = content[pos]
        if ch == "\\":
            pos += 2
            continue
        if ch == '"':
            return pos + 1
        pos += 1
    return pos


def _skip_ws(content: str, pos: int, length: int) -> int:
    while pos < length and content[pos] in " \t\r\n":
        pos += 1
    return pos


def _skip_to_value_start(
    content: str,
    pos: int,
    length: int,
    current_line: int,
) -> tuple[int, int]:
    while pos < length and content[pos] != ":":
        if content[pos] == "\n":
            current_line += 1
        pos += 1
    if pos >= length:
        return pos, current_line
    pos += 1

    while pos < length and content[pos] in " \t\r\n":
        if content[pos] == "\n":
            current_line += 1
        pos += 1

    return pos, current_line


def _scan_container_end(content: str, pos: int, length: int, current_line: int) -> int:
    depth = 1
    pos += 1
    while pos < length and depth > 0:
        c = content[pos]
        if c == "\n":
            current_line += 1
        elif c == '"':
            pos = _skip_string(content, pos + 1)
            continue
        elif c in "{[":
            depth += 1
        elif c in "}]":
            depth -= 1
        pos += 1
    return current_line



import types
from typing import TYPE_CHECKING, Annotated, Literal, Self
from urllib.parse import ParseResult

if TYPE_CHECKING:
    from dature.field_path import FieldPath
    from dature.protocols import ValidatorProtocol

type JSONValue = dict[str, JSONValue] | list[JSONValue] | str | int | float | bool | None


class NotLoaded:
    _instance: "NotLoaded | None" = None

    def __new__(cls) -> Self:
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance  # type: ignore[return-value]

    def __repr__(self) -> str:
        return "NOT_LOADED"

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return hash("NOT_LOADED")


NOT_LOADED = NotLoaded()

type ProbeValue = dict[str, ProbeValue] | list[ProbeValue] | str | int | float | bool | NotLoaded | None

type ProbeDict = dict[str, ProbeValue]

# Result of get_type_hints() / get_args(): concrete class or parameterized generic
type TypeAnnotation = type[object] | types.GenericAlias

# Examples: "app", "app.database", "app.database.host"
type DotSeparatedPath = Annotated[str, "Dot-separated path for nested dictionary navigation"]

type NameStyle = Literal[
    "lower_snake",
    "upper_snake",
    "lower_camel",
    "upper_camel",
    "lower_kebab",
    "upper_kebab",
]

# Keys are FieldPath at runtime, but F[Type].field returns the field's static type (str, int, etc.)
# due to the overload trick for IDE autocompletion, so we accept those types here too.
type _FieldMappingKey = "FieldPath | str | int | float | bool | None"
type FieldMapping = dict[_FieldMappingKey, str | tuple[str, ...]]

type URL = ParseResult

type Base64UrlBytes = bytes
type Base64UrlStr = str

type ExpandEnvVarsMode = Literal["disabled", "default", "empty", "strict"]

type _ValidatorKey = "FieldPath | str | int | float | bool | None"
type FieldValidators = dict[_ValidatorKey, "ValidatorProtocol | tuple[ValidatorProtocol, ...]"]



from dataclasses import dataclass

from dature.path_finders.char_base import (
    CharPathFinder,
    KeyParseResult,
    PosLine,
    StackEntry,
    build_path,
    increment_parent_array,
)


class Json5PathFinder(CharPathFinder):
    def _skip_noise(self, content: str, pos: int, line: int, length: int) -> PosLine | None:
        ch = content[pos]
        if ch in " \t\r\n":
            if ch == "\n":
                line += 1
            return PosLine(pos=pos + 1, line=line)

        if ch == ",":
            return PosLine(pos=pos + 1, line=line)

        if ch != "/" or pos + 1 >= length:
            return None

        next_ch = content[pos + 1]
        if next_ch == "/":
            return PosLine(pos=_skip_line_comment(content, pos + 2, length), line=line)
        if next_ch == "*":
            block = _skip_block_comment(content, pos + 2, length, line)
            return PosLine(pos=block.pos, line=block.line)

        return None

    def _try_parse_key(
        self,
        content: str,
        pos: int,
        line: int,
        length: int,
        stack: list[StackEntry],
        target_path: list[str],
    ) -> KeyParseResult | None:
        ch = content[pos]

        if ch in {'"', "'"}:
            return _handle_quoted(content, pos, line, length, stack, target_path)

        if _is_ident_start(ch):
            return _handle_identifier(content, pos, line, length, stack, target_path)

        return None

    def _find_value_end_line(self, content: str, pos: int, length: int, start_line: int) -> int:
        current_line = start_line

        pl = _skip_ws_and_comments(content, pos, length, current_line)
        pos = pl.pos
        current_line = pl.line
        if pos >= length or content[pos] != ":":
            return start_line
        pos += 1

        pl = _skip_ws_and_comments(content, pos, length, current_line)
        pos = pl.pos
        current_line = pl.line

        if pos >= length:
            return current_line

        ch = content[pos]

        if ch in {'"', "'"}:
            end = _skip_quoted_string(content, pos + 1, ch, current_line)
            return end.line

        if ch in "{[":
            return _scan_json5_container_end(content, pos, length, current_line)

        return current_line


@dataclass(frozen=True, slots=True)
class _PosLine:
    pos: int
    line: int


def _handle_quoted(
    content: str,
    pos: int,
    line: int,
    length: int,
    stack: list[StackEntry],
    target_path: list[str],
) -> KeyParseResult:
    quote = content[pos]
    string_start = pos + 1
    end = _skip_quoted_string(content, string_start, quote, line)
    string_value = content[string_start : end.pos - 1]

    colon = _skip_ws_and_comments(content, end.pos, length, end.line)

    if colon.pos >= length or content[colon.pos] != ":":
        increment_parent_array(stack)
        return KeyParseResult(pos=end.pos, line=end.line, last_key=None, found=False)

    if build_path(stack, string_value) == target_path:
        return KeyParseResult(pos=end.pos, line=line, last_key=None, found=True)

    return KeyParseResult(pos=colon.pos + 1, line=colon.line, last_key=string_value, found=False)


def _handle_identifier(
    content: str,
    pos: int,
    line: int,
    length: int,
    stack: list[StackEntry],
    target_path: list[str],
) -> KeyParseResult:
    ident_end = _skip_identifier(content, pos, length)
    ident = content[pos:ident_end]

    colon = _skip_ws_and_comments(content, ident_end, length, line)

    if colon.pos >= length or content[colon.pos] != ":":
        increment_parent_array(stack)
        return KeyParseResult(pos=ident_end, line=line, last_key=None, found=False)

    if build_path(stack, ident) == target_path:
        return KeyParseResult(pos=ident_end, line=line, last_key=None, found=True)

    return KeyParseResult(pos=colon.pos + 1, line=colon.line, last_key=ident, found=False)


def _skip_quoted_string(
    content: str,
    pos: int,
    quote: str,
    line: int,
) -> _PosLine:
    """Advance past a quoted string body. Returns position after the closing quote."""
    length = len(content)
    while pos < length:
        ch = content[pos]
        if ch == "\\":
            if pos + 1 < length and content[pos + 1] == "\n":
                line += 1
            pos += 2
            continue
        if ch == "\n":
            line += 1
            pos += 1
            continue
        if ch == quote:
            return _PosLine(pos=pos + 1, line=line)
        pos += 1
    return _PosLine(pos=pos, line=line)


def _is_ident_start(ch: str) -> bool:
    return ch.isalpha() or ch in {"_", "$"}


def _is_ident_char(ch: str) -> bool:
    return ch.isalnum() or ch in {"_", "$"}


def _skip_identifier(content: str, pos: int, length: int) -> int:
    while pos < length and _is_ident_char(content[pos]):
        pos += 1
    return pos


def _skip_line_comment(content: str, pos: int, length: int) -> int:
    while pos < length and content[pos] != "\n":
        pos += 1
    return pos


def _skip_block_comment(
    content: str,
    pos: int,
    length: int,
    line: int,
) -> _PosLine:
    while pos < length:
        if content[pos] == "\n":
            line += 1
        elif content[pos] == "*" and pos + 1 < length and content[pos + 1] == "/":
            return _PosLine(pos=pos + 2, line=line)
        pos += 1
    return _PosLine(pos=pos, line=line)


def _skip_ws_and_comments(
    content: str,
    pos: int,
    length: int,
    line: int,
) -> _PosLine:
    while pos < length:
        ch = content[pos]
        if ch in " \t\r":
            pos += 1
            continue
        if ch == "\n":
            line += 1
            pos += 1
            continue
        if ch == "/" and pos + 1 < length and content[pos + 1] == "/":
            pos = _skip_line_comment(content, pos + 2, length)
            continue
        if ch == "/" and pos + 1 < length and content[pos + 1] == "*":
            block = _skip_block_comment(content, pos + 2, length, line)
            pos = block.pos
            line = block.line
            continue
        return _PosLine(pos=pos, line=line)
    return _PosLine(pos=pos, line=line)


def _scan_json5_container_end(
    content: str,
    pos: int,
    length: int,
    current_line: int,
) -> int:
    depth = 1
    pos += 1
    while pos < length and depth > 0:
        c = content[pos]
        if c == "\n":
            current_line += 1
        elif c in {'"', "'"}:
            end = _skip_quoted_string(content, pos + 1, c, current_line)
            pos = end.pos
            current_line = end.line
            continue
        elif c == "/" and pos + 1 < length:
            nc = content[pos + 1]
            if nc == "/":
                pos = _skip_line_comment(content, pos + 2, length)
                continue
            if nc == "*":
                block = _skip_block_comment(content, pos + 2, length, current_line)
                pos = block.pos
                current_line = block.line
                continue
        elif c in "{[":
            depth += 1
        elif c in "}]":
            depth -= 1
        pos += 1
    return current_line



import os
from datetime import date, datetime, time
from pathlib import Path
from typing import Any, cast

from adaptix import loader
from adaptix.provider import Provider

from dature.env_expand import expand_env_vars
from dature.protocols import ValidatorProtocol
from dature.sources_loader.base import BaseLoader
from dature.sources_loader.loaders import (
    bool_loader,
    bytearray_from_json_string,
    date_from_string,
    datetime_from_string,
    none_from_empty_string,
    optional_from_empty_string,
    time_from_string,
)
from dature.types import DotSeparatedPath, ExpandEnvVarsMode, FieldMapping, FieldValidators, JSONValue, NameStyle


def _set_nested(d: dict[Any, Any], keys: list[str], value: str) -> None:
    for key in keys[:-1]:
        d = d.setdefault(key, {})
    d[keys[-1]] = value


class EnvLoader(BaseLoader):
    display_name = "env"

    def __init__(  # noqa: PLR0913
        self,
        *,
        prefix: DotSeparatedPath | None = None,
        split_symbols: str = "__",
        name_style: NameStyle | None = None,
        field_mapping: FieldMapping | None = None,
        root_validators: tuple[ValidatorProtocol, ...] | None = None,
        validators: FieldValidators | None = None,
        expand_env_vars: ExpandEnvVarsMode = "default",
    ) -> None:
        self._split_symbols = split_symbols
        super().__init__(
            prefix=prefix,
            name_style=name_style,
            field_mapping=field_mapping,
            root_validators=root_validators,
            validators=validators,
            expand_env_vars=expand_env_vars,
        )

    def _additional_loaders(self) -> list[Provider]:
        return [
            loader(date, date_from_string),
            loader(datetime, datetime_from_string),
            loader(time, time_from_string),
            loader(bytearray, bytearray_from_json_string),
            loader(type(None), none_from_empty_string),
            loader(str | None, optional_from_empty_string),
            loader(bool, bool_loader),
        ]

    def _load(self, _: Path) -> JSONValue:
        return cast("JSONValue", os.environ)

    def _pre_processing(self, data: JSONValue) -> JSONValue:
        data_dict = cast("dict[str, str]", data)
        result: dict[str, JSONValue] = {}

        for key, value in data_dict.items():
            self._pre_processed_row(key=key, value=value, result=result)

        expanded = expand_env_vars(result, mode=self._expand_env_vars_mode)
        return self._parse_string_values(expanded)

    def _pre_processed_row(self, key: str, value: str, result: dict[str, JSONValue]) -> None:
        if self._prefix and not key.startswith(self._prefix):
            return

        processed_key = key[len(self._prefix) :] if self._prefix else key
        processed_key = processed_key.lower()

        parts = processed_key.split(self._split_symbols)
        if len(parts) > 1:
            _set_nested(result, parts, value)
        else:
            result[processed_key] = value


class EnvFileLoader(EnvLoader):
    display_name = "envfile"

    def _load(self, path: Path) -> JSONValue:
        env_vars: dict[str, JSONValue] = {}

        with path.open() as file_:
            for raw_line in file_:
                if not (line := raw_line.strip()) or line.startswith("#"):
                    continue

                if "=" not in line:
                    continue

                key, value = line.split("=", 1)
                key = key.strip()
                value = value.strip().strip('"').strip("'")
                self._pre_processed_row(key=key, value=value, result=env_vars)

        return env_vars

    def _pre_processing(self, data: JSONValue) -> JSONValue:
        expanded = expand_env_vars(data, mode=self._expand_env_vars_mode)
        return self._parse_string_values(expanded)



from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from typing import Annotated, Literal

import pytest

from dature import LoadMetadata, load
from dature.errors import DatureConfigError
from dature.sources_loader.yaml_ import Yaml11Loader, Yaml12Loader
from dature.validators.number import Ge, Le
from dature.validators.sequence import MinItems, UniqueItems
from dature.validators.string import MaxLength, MinLength, RegexPattern


@dataclass
class Address:
    city: Annotated[str, MinLength(value=2)]
    zip_code: Annotated[str, RegexPattern(pattern=r"^\d{5}$")]


@dataclass
class ErrorConfig:
    port: int
    host: str
    status: Literal["active", "inactive"]
    name: Annotated[str, MinLength(value=3), MaxLength(value=50)]
    email: Annotated[str, RegexPattern(pattern=r"^[\w.-]+@[\w.-]+\.\w+$")]
    age: Annotated[int, Ge(value=0), Le(value=150)]
    tags: Annotated[list[str], MinItems(value=1), UniqueItems()]
    address: Address


@dataclass
class LoadErrorConfig:
    port: int
    host: str
    status: Literal["active", "inactive"]
    name: str
    email: str
    age: int
    tags: list[str]
    address: Address


@dataclass
class ValidationErrorConfig:
    name: Annotated[str, MinLength(value=3), MaxLength(value=50)]
    email: Annotated[str, RegexPattern(pattern=r"^[\w.-]+@[\w.-]+\.\w+$")]
    age: Annotated[int, Ge(value=0), Le(value=150)]
    tags: Annotated[list[str], MinItems(value=1), UniqueItems()]
    address: Address


FIXTURES_DIR = Path(__file__).parent / "fixtures"

ALL_SOURCES = [
    ("errors.json", {}),
    ("errors.json5", {}),
    ("errors.yaml", {}),
    ("errors.yaml", {"loader": Yaml11Loader}),
    ("errors.yaml", {"loader": Yaml12Loader}),
    ("errors.toml", {}),
    ("errors.ini", {"prefix": "config"}),
    ("errors.env", {}),
]


def _load_error_message(p: str) -> str:
    return dedent("""\
        LoadErrorConfig loading errors (5)

          [port]  Bad string format
           └── FILE '{path}', line 2
               "port": "abc",

          [host]  Missing required field
           └── FILE '{path}'

          [status]  Invalid variant: 'unknown'
           └── FILE '{path}', line 3
               "status": "unknown",

          [address.city]  Value must have at least 2 characters
           └── FILE '{path}', line 9
               "city": "N",

          [address.zip_code]  Value must match pattern '^\\d{5}$'
           └── FILE '{path}', line 10
               "zip_code": "ABCDE"
        """).replace("{path}", p)


def _validation_error_message(p: str) -> str:
    return dedent("""\
        ValidationErrorConfig loading errors (6)

          [name]  Value must have at least 3 characters
           └── FILE '{path}', line 4
               "name": "AB",

          [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
           └── FILE '{path}', line 5
               "email": "not-an-email",

          [age]  Value must be less than or equal to 150
           └── FILE '{path}', line 6
               "age": 200,

          [tags]  Value must have at least 1 items
           └── FILE '{path}', line 7
               "tags": [],

          [address.city]  Value must have at least 2 characters
           └── FILE '{path}', line 9
               "city": "N",

          [address.zip_code]  Value must match pattern '^\\d{5}$'
           └── FILE '{path}', line 10
               "zip_code": "ABCDE"
        """).replace("{path}", p)


EXPECTED_LOAD_MESSAGES: dict[str, str] = {}
EXPECTED_VALIDATION_MESSAGES: dict[str, str] = {}


def _file(name: str) -> str:
    return str(FIXTURES_DIR / name)


# --- json ---
EXPECTED_LOAD_MESSAGES["json"] = _load_error_message(_file("errors.json"))
EXPECTED_VALIDATION_MESSAGES["json"] = _validation_error_message(_file("errors.json"))

# --- json5 ---
_json5_path = _file("errors.json5")
EXPECTED_LOAD_MESSAGES["json5"] = dedent(f"""\
    LoadErrorConfig loading errors (5)

      [port]  Bad string format
       └── FILE '{_json5_path}', line 3
           port: "abc",

      [host]  Missing required field
       └── FILE '{_json5_path}'

      [status]  Invalid variant: 'unknown'
       └── FILE '{_json5_path}', line 5
           status: "unknown",

      [address.city]  Value must have at least 2 characters
       └── FILE '{_json5_path}', line 16
           city: "N",

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_json5_path}', line 17
           zip_code: "ABCDE",
    """)

EXPECTED_VALIDATION_MESSAGES["json5"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── FILE '{_json5_path}', line 7
           name: "AB",

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── FILE '{_json5_path}', line 9
           email: "not-an-email",

      [age]  Value must be less than or equal to 150
       └── FILE '{_json5_path}', line 11
           age: 200,

      [tags]  Value must have at least 1 items
       └── FILE '{_json5_path}', line 13
           tags: [],

      [address.city]  Value must have at least 2 characters
       └── FILE '{_json5_path}', line 16
           city: "N",

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_json5_path}', line 17
           zip_code: "ABCDE",
    """)

# --- yaml ---
_yaml_path = _file("errors.yaml")
EXPECTED_LOAD_MESSAGES["yaml"] = dedent(f"""\
    LoadErrorConfig loading errors (5)

      [port]  Bad string format
       └── FILE '{_yaml_path}', line 1
           port: "abc"

      [host]  Missing required field
       └── FILE '{_yaml_path}'

      [status]  Invalid variant: 'unknown'
       └── FILE '{_yaml_path}', line 2
           status: "unknown"

      [address.city]  Value must have at least 2 characters
       └── FILE '{_yaml_path}', line 8
           city: "N"

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_yaml_path}', line 9
           zip_code: "ABCDE"
    """)
EXPECTED_VALIDATION_MESSAGES["yaml"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── FILE '{_yaml_path}', line 3
           name: "AB"

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── FILE '{_yaml_path}', line 4
           email: "not-an-email"

      [age]  Value must be less than or equal to 150
       └── FILE '{_yaml_path}', line 5
           age: 200

      [tags]  Value must have at least 1 items
       └── FILE '{_yaml_path}', line 6
           tags: []

      [address.city]  Value must have at least 2 characters
       └── FILE '{_yaml_path}', line 8
           city: "N"

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_yaml_path}', line 9
           zip_code: "ABCDE"
    """)

# --- toml ---
_toml_path = _file("errors.toml")
EXPECTED_LOAD_MESSAGES["toml"] = dedent(f"""\
    LoadErrorConfig loading errors (5)

      [port]  Bad string format
       └── FILE '{_toml_path}', line 1
           port = "abc"

      [host]  Missing required field
       └── FILE '{_toml_path}'

      [status]  Invalid variant: 'unknown'
       └── FILE '{_toml_path}', line 2
           status = "unknown"

      [address.city]  Value must have at least 2 characters
       └── FILE '{_toml_path}', line 9
           city = "N"

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_toml_path}', line 10
           zip_code = "ABCDE"
    """)
EXPECTED_VALIDATION_MESSAGES["toml"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── FILE '{_toml_path}', line 3
           name = "AB"

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── FILE '{_toml_path}', line 4
           email = "not-an-email"

      [age]  Value must be less than or equal to 150
       └── FILE '{_toml_path}', line 5
           age = 200

      [tags]  Value must have at least 1 items
       └── FILE '{_toml_path}', line 6
           tags = []

      [address.city]  Value must have at least 2 characters
       └── FILE '{_toml_path}', line 9
           city = "N"

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_toml_path}', line 10
           zip_code = "ABCDE"
    """)

# --- ini ---
_ini_path = _file("errors.ini")
EXPECTED_LOAD_MESSAGES["ini"] = dedent(f"""\
    LoadErrorConfig loading errors (5)

      [port]  Bad string format
       └── FILE '{_ini_path}', line 2
           port = abc

      [host]  Missing required field
       └── FILE '{_ini_path}'

      [status]  Invalid variant: 'unknown'
       └── FILE '{_ini_path}', line 3
           status = unknown

      [address.city]  Value must have at least 2 characters
       └── FILE '{_ini_path}', line 10
           city = N

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_ini_path}', line 11
           zip_code = ABCDE
    """)
EXPECTED_VALIDATION_MESSAGES["ini"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── FILE '{_ini_path}', line 4
           name = AB

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── FILE '{_ini_path}', line 5
           email = not-an-email

      [age]  Value must be less than or equal to 150
       └── FILE '{_ini_path}', line 6
           age = 200

      [tags]  Value must have at least 1 items
       └── FILE '{_ini_path}', line 7
           tags = []

      [address.city]  Value must have at least 2 characters
       └── FILE '{_ini_path}', line 10
           city = N

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── FILE '{_ini_path}', line 11
           zip_code = ABCDE
    """)

# --- env ---
_env_path = _file("errors.env")
EXPECTED_LOAD_MESSAGES["env"] = dedent(f"""\
    LoadErrorConfig loading errors (5)

      [port]  Bad string format
       └── ENV FILE '{_env_path}', var 'PORT'
           PORT=abc

      [host]  Missing required field
       └── ENV FILE '{_env_path}', var 'HOST'

      [status]  Invalid variant: 'unknown'
       └── ENV FILE '{_env_path}', var 'STATUS'
           STATUS=unknown

      [address.city]  Value must have at least 2 characters
       └── ENV FILE '{_env_path}', var 'ADDRESS__CITY'
           ADDRESS__CITY=N

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── ENV FILE '{_env_path}', var 'ADDRESS__ZIP_CODE'
           ADDRESS__ZIP_CODE=ABCDE
    """)
EXPECTED_VALIDATION_MESSAGES["env"] = dedent(f"""\
    ValidationErrorConfig loading errors (6)

      [name]  Value must have at least 3 characters
       └── ENV FILE '{_env_path}', var 'NAME'
           NAME=AB

      [email]  Value must match pattern '^[\\w.-]+@[\\w.-]+\\.\\w+$'
       └── ENV FILE '{_env_path}', var 'EMAIL'
           EMAIL=not-an-email

      [age]  Value must be less than or equal to 150
       └── ENV FILE '{_env_path}', var 'AGE'
           AGE=200

      [tags]  Value must have at least 1 items
       └── ENV FILE '{_env_path}', var 'TAGS'
           TAGS=[]

      [address.city]  Value must have at least 2 characters
       └── ENV FILE '{_env_path}', var 'ADDRESS__CITY'
           ADDRESS__CITY=N

      [address.zip_code]  Value must match pattern '^\\d{{5}}$'
       └── ENV FILE '{_env_path}', var 'ADDRESS__ZIP_CODE'
           ADDRESS__ZIP_CODE=ABCDE
    """)


def _get_format_key(fixture_file: str) -> str:
    suffix = Path(fixture_file).suffix.lstrip(".")
    if suffix == "json5":
        return "json5"
    return suffix


@pytest.mark.parametrize(("fixture_file", "metadata_kwargs"), ALL_SOURCES)
def test_load_error_types(
    fixture_file: str,
    metadata_kwargs: dict[str, str],
) -> None:
    path = str(FIXTURES_DIR / fixture_file)
    metadata = LoadMetadata(file_=path, **metadata_kwargs)

    with pytest.raises(DatureConfigError) as exc_info:
        load(metadata, LoadErrorConfig)

    fmt = _get_format_key(fixture_file)
    assert str(exc_info.value) == EXPECTED_LOAD_MESSAGES[fmt]


@pytest.mark.parametrize(("fixture_file", "metadata_kwargs"), ALL_SOURCES)
def test_validation_error_types(
    fixture_file: str,
    metadata_kwargs: dict[str, str],
) -> None:
    path = str(FIXTURES_DIR / fixture_file)
    metadata = LoadMetadata(file_=path, **metadata_kwargs)

    with pytest.raises(DatureConfigError) as exc_info:
        load(metadata, ValidationErrorConfig)

    fmt = _get_format_key(fixture_file)
    assert str(exc_info.value) == EXPECTED_VALIDATION_MESSAGES[fmt]
